"""Configuration for dynamic_rest."""
import inspect

from django.conf import settings as django_settings
from django.test.signals import setting_changed

DYNAMIC_REST = {
    # DEBUG: enable/disable internal debugging
    "DEBUG": False,
    # ENABLE_BROWSABLE_API: enable/disable the browsable API.
    # It can be useful to disable it in production.
    "ENABLE_BROWSABLE_API": True,
    # ENABLE_LINKS: enable/disable relationship links
    "ENABLE_LINKS": True,
    # ENABLE_SERIALIZER_CACHE: enable/disable caching of related serializers
    "ENABLE_SERIALIZER_CACHE": True,
    # ENABLE_SERIALIZER_OBJECT_CACHE: enable/disable caching of serialized
    # objects within a serializer instance/context. This can yield
    # significant performance improvements in cases where the same objects
    # are sideloaded repeatedly.
    "ENABLE_SERIALIZER_OBJECT_CACHE": True,
    # ENABLE_SERIALIZER_OPTIMIZATIONS: enable/disable representation speedups
    "ENABLE_SERIALIZER_OPTIMIZATIONS": True,
    # ENABLE_BULK_PARTIAL_CREATION: enable/disable partial creation in bulk
    "ENABLE_BULK_PARTIAL_CREATION": False,
    # ENABLE_BULK_UPDATE: enable/disable update in bulk
    "ENABLE_BULK_UPDATE": True,
    # ENABLE_PATCH_ALL: enable/disable patch by queryset
    "ENABLE_PATCH_ALL": False,
    # DEFER_MANY_RELATIONS: automatically defer many-relations, unless
    # `deferred=False` is explicitly set on the field.
    "DEFER_MANY_RELATIONS": False,
    # LIST_SERIALIZER_CLASS: Globally override the list serializer class.
    # Default is `DynamicListSerializer` and also can be overridden for
    # each serializer class by setting `Meta.list_serializer_class`.
    "LIST_SERIALIZER_CLASS": None,
    # MAX_PAGE_SIZE: global setting for max page size.
    # Can be overriden at the viewset level.
    "MAX_PAGE_SIZE": None,
    # PAGE_QUERY_PARAM: global setting for the pagination query parameter.
    # Can be overriden at the viewset level.
    "PAGE_QUERY_PARAM": "page",
    # PAGE_SIZE: global setting for page size.
    # Can be overriden at the viewset level.
    "PAGE_SIZE": None,
    # PAGE_SIZE_QUERY_PARAM: global setting for the page size query parameter.
    # Can be overriden at the viewset level.
    "PAGE_SIZE_QUERY_PARAM": "per_page",
    # EXCLUDE_COUNT_QUERY_PARAM: global setting for the query parameter
    # that disables counting during PageNumber pagination
    "EXCLUDE_COUNT_QUERY_PARAM": "exclude_count",
    # ADDITIONAL_PRIMARY_RESOURCE_PREFIX: String to prefix additional
    # instances of the primary resource when sideloading.
    "ADDITIONAL_PRIMARY_RESOURCE_PREFIX": "+",
    # Enables host-relative links.  Only compatible with resources registered
    # through the dynamic router.  If a resource doesn't have a canonical
    # path registered, links will default back to being resource-relative urls
    "ENABLE_HOST_RELATIVE_LINKS": True,
    # Enables caching of serializer fields to speed up serializer usage
    # Needs to also be configured on a per-serializer basis
    "ENABLE_FIELDS_CACHE": False,
    # Enables use of hashid fields
    "ENABLE_HASHID_FIELDS": False,
    # Salt value to salt hash ids.
    # Needs to be non-nullable if 'ENABLE_HASHID_FIELDS' is set to True
    "HASHIDS_SALT": None,
}


# Attributes where the value should be a class (or path to a class)
CLASS_ATTRS = [
    "LIST_SERIALIZER_CLASS",
]


class Settings(object):
    """Settings class for dynamic rest."""

    def __init__(self, name, defaults, existing_settings, class_attrs=None):
        """Initialise the settings."""
        self.name = name
        self.defaults = defaults
        self.keys = set(defaults.keys())
        self.class_attrs = class_attrs

        self._cache = {}
        self._reload(getattr(existing_settings, self.name, {}))

        setting_changed.connect(self._settings_changed)

    def _reload(self, value):
        """Reload settings after a change."""
        self.settings = value
        self._cache = {}

    def _load_class(self, attr, val):
        """Load a class from a string."""
        if inspect.isclass(val):
            return val
        elif isinstance(val, str):
            parts = val.split(".")
            module_path = ".".join(parts[:-1])
            class_name = parts[-1]
            mod = __import__(module_path, fromlist=[class_name])
            return getattr(mod, class_name)
        elif val:
            raise ValueError(f"{attr} must be string or a class")

    def __getattr__(self, attr):
        """Get a setting."""
        if attr not in self._cache:
            if attr not in self.keys:
                raise AttributeError(f"Invalid API setting: '{attr}'")

            if attr in self.settings:
                val = self.settings[attr]
            else:
                val = self.defaults[attr]

            if attr in self.class_attrs and val:
                val = self._load_class(attr, val)

            # Cache the result
            self._cache[attr] = val

        return self._cache[attr]

    def _settings_changed(self, *_, **kwargs):
        """Handle changes to core settings."""
        setting, value = kwargs["setting"], kwargs["value"]
        if setting == self.name:
            self._reload(value)


settings = Settings("DYNAMIC_REST", DYNAMIC_REST, django_settings, CLASS_ATTRS)



"""This module contains custom pagination classes."""
from collections import OrderedDict

from django.core.paginator import InvalidPage
from django.utils.functional import cached_property
from rest_framework.exceptions import NotFound
from rest_framework.pagination import PageNumberPagination
from rest_framework.response import Response
from rest_framework.settings import api_settings

from dynamic_rest.conf import settings
from dynamic_rest.paginator import DynamicPaginator


class DynamicPageNumberPagination(PageNumberPagination):
    """A subclass of PageNumberPagination.

    Adds support for pagination metadata and overrides for
    pagination query parameters.
    """

    page_size_query_param = settings.PAGE_SIZE_QUERY_PARAM
    exclude_count_query_param = settings.EXCLUDE_COUNT_QUERY_PARAM
    page_query_param = settings.PAGE_QUERY_PARAM
    max_page_size = settings.MAX_PAGE_SIZE
    page_size = settings.PAGE_SIZE or api_settings.PAGE_SIZE
    django_paginator_class = DynamicPaginator

    def get_page_metadata(self):
        """Return metadata about the current page."""
        # always returns page, per_page
        # also returns total_results and total_pages
        # (unless EXCLUDE_COUNT_QUERY_PARAM is set)
        page = self.page
        meta = {"page": page.number, "per_page": self.get_page_size(self.request)}
        if not self.exclude_count:
            meta["total_results"] = page.paginator.count
            meta["total_pages"] = page.paginator.num_pages
        else:
            meta["more_pages"] = self.more_pages
        return meta

    def get_paginated_response(self, data):
        """Return a paginated response."""
        meta = self.get_page_metadata()
        if isinstance(data, list):
            result = OrderedDict()
            if not self.exclude_count:
                result["count"] = self.page.paginator.count
                result["next"] = self.get_next_link()
                result["previous"] = self.get_previous_link()
            result["results"] = data
            result["meta"] = meta
        else:
            result = data
            if "meta" in result:
                result["meta"].update(meta)
            else:
                result["meta"] = meta
        return Response(result)

    @cached_property
    def exclude_count(self):
        """Return whether to exclude count."""
        return self.request.query_params.get(self.exclude_count_query_param)

    def get_page_number(self, request, paginator):
        """Return the page number."""
        page_number = request.query_params.get(self.page_query_param, 1)
        if page_number in self.last_page_strings:
            page_number = paginator.num_pages
        return page_number

    def paginate_queryset(self, queryset, request, **__):
        """Paginate a queryset.

        If required, either returning a page object,
        or `None` if pagination is not configured for this view.
        """
        if "exclude_count" in self.__dict__:
            self.__dict__.pop("exclude_count")

        page_size = self.get_page_size(request)
        if not page_size:
            return None
        self.request = request
        exclude = self.exclude_count
        paginator = self.django_paginator_class(
            queryset, page_size, exclude_count=exclude
        )
        page_number = self.get_page_number(request, paginator)

        try:
            self.page = paginator.page(page_number)
        except InvalidPage as exc:
            msg = self.invalid_page_message.format(
                page_number=page_number, message=str(exc)
            )
            raise NotFound(msg) from exc

        if paginator.num_pages > 1 and self.template is not None:
            # The browsable API should display pagination controls.
            self.display_page_controls = True

        result = list(self.page)
        if exclude:
            if len(result) > page_size:
                # if exclude_count is set, we fetch one extra item
                result = result[:page_size]
                self.more_pages = True
            else:
                self.more_pages = False
        return result



"""Module containing Django meta helpers."""
from __future__ import annotations

from functools import lru_cache
from itertools import chain

from django.core.exceptions import FieldDoesNotExist
from django.db.models import ManyToOneRel  # tested in 1.9
from django.db.models import OneToOneRel  # tested in 1.9
from django.db.models import (
    ForeignKey,
    ManyToManyField,
    ManyToManyRel,
    Model,
    OneToOneField,
)

from dynamic_rest.related import RelatedObject


def is_model_field(model, field_name):
    """Check whether a given field exists on a model.

    Arguments:
        model: a Django model
        field_name: the name of a field

    Returns:
        True if `field_name` exists on `model`, False otherwise.
    """
    try:
        get_model_field(model, field_name)
        return True
    except AttributeError:
        return False


@lru_cache()
def get_model_relationships(meta) -> dict:
    """Return a dictionary of all relationships on a model."""
    related_objs = (
        f
        for f in meta.get_fields()
        if (f.one_to_many or f.one_to_one) and f.auto_created and not f.concrete
    )
    related_m2m_objs = (
        f
        for f in meta.get_fields(include_hidden=True)
        if f.many_to_many and f.auto_created
    )
    return {o.get_accessor_name(): o for o in chain(related_objs, related_m2m_objs)}


@lru_cache()
def get_virtual_fields(meta) -> dict | None:
    """Return a dictionary of all virtual fields on a model."""
    if hasattr(meta, "virtual_fields"):
        return {f.name: f for f in meta.virtual_fields}


def get_model_field(model: Model | None, field_name):
    """Return a field given a model and field name.

    Arguments:
        model: a Django model
        field_name: the name of a field

    Returns:
        A Django field if `field_name` is a valid field for `model`,
            None otherwise.
    """
    if not model:
        # TODO: This might be a bug.
        return None

    meta = model._meta  # pylint: disable=protected-access
    try:
        return meta.get_field(field_name)
    except FieldDoesNotExist as exc:
        meta._related_fields_cache = (  # pylint: disable=protected-access
            cache
        ) = getattr(meta, "_related_fields_cache", get_model_relationships(meta))

        if field_name in cache:
            return cache[field_name]
        # check virtual fields (1.7)
        if virtual_fields := get_virtual_fields(meta):
            if field := virtual_fields.get(field_name):
                return field

        raise AttributeError(f"{field_name} is not a valid field for {model}") from exc


def get_model_field_and_type(model, field_name):
    """Return a field and its type given a model and field name."""
    field = get_model_field(model, field_name)

    # Django 1.7 (and 1.8?)
    if isinstance(field, RelatedObject):
        if isinstance(field.field, OneToOneField):
            return field, "o2or"
        elif isinstance(field.field, ManyToManyField):
            return field, "m2m"
        elif isinstance(field.field, ForeignKey):
            return field, "m2o"
        else:
            raise RuntimeError("Unexpected field type")

    # Django 1.9
    type_map = [
        (OneToOneField, "o2o"),
        (OneToOneRel, "o2or"),  # is subclass of m2o so check first
        (ManyToManyField, "m2m"),
        (ManyToOneRel, "m2o"),
        (ManyToManyRel, "m2m"),
        (ForeignKey, "fk"),  # check last
    ]
    for cls, type_str in type_map:
        if isinstance(field, cls):
            return (
                field,
                type_str,
            )

    return (
        field,
        "",
    )


def is_field_remote(model, field_name):
    """Check whether a given model field is a remote field.

    A remote field is the inverse of a one-to-many or a
    many-to-many relationship.

    Arguments:
        model: a Django model
        field_name: the name of a field

    Returns:
        True if `field_name` is a remote field, False otherwise.
    """
    if not hasattr(model, "_meta"):
        # ephemeral model with no metaclass
        return False

    model_field = get_model_field(model, field_name)
    return isinstance(model_field, (ManyToManyField, RelatedObject))


def get_model_table(model):
    """Return the table name for a given model."""
    return model._meta.db_table  # pylint: disable=protected-access



"""This module contains custom ViewSet classes."""
from __future__ import annotations

import json
import logging
from typing import Protocol, runtime_checkable

from django.core.exceptions import ObjectDoesNotExist
from django.db import IntegrityError, transaction
from django.http import QueryDict
from rest_framework import exceptions, status, viewsets
from rest_framework.exceptions import ValidationError
from rest_framework.renderers import BaseRenderer, BrowsableAPIRenderer
from rest_framework.request import Request
from rest_framework.response import Response

from dynamic_rest.conf import settings
from dynamic_rest.filters import DynamicFilterBackend, DynamicSortingFilter
from dynamic_rest.metadata import DynamicMetadata
from dynamic_rest.pagination import DynamicPageNumberPagination
from dynamic_rest.processors import SideloadingProcessor
from dynamic_rest.utils import is_truthy

UPDATE_REQUEST_METHODS = ("PUT", "PATCH", "POST")
DELETE_REQUEST_METHOD = "DELETE"
PATCH = "PATCH"

logger = logging.getLogger(__name__)


def _extract_object_params(
    request: Request, name: str, raw: bool = False
) -> dict | None:
    """Extract object params, return as dict.

    Always returns None if raw??? # TODO: Make sure this makes sense.

    Args:
        request: Request object.
        name: Name of the object.
        raw: Raw flag.

    Returns:
        dict | None: Parsed object params.
    """
    params = request.query_params.lists()
    logger.debug("Extracting object params: %s", name)
    params_map = {}
    original_name = name
    prefix = name[:-1]
    offset = len(prefix)
    logger.debug("Name: %s, Prefix: %s, Offset: %s", name, prefix, offset)

    for param_name, value in params:
        if param_name == original_name:
            if raw and value:
                # filter{} as object
                return json.loads(value[0])
            else:
                continue

        if not param_name.startswith(prefix):
            continue

        if param_name.endswith("}"):
            param_name = param_name[offset:-1]
        elif param_name.endswith("}[]"):
            # strip off trailing }[]
            # this fixes an Ember queryparams issue
            param_name = param_name[offset:-3]
        else:
            # malformed argument like:
            # filter{foo=bar
            raise exceptions.ParseError(
                f'"{param_name}" is not a well-formed filter key.'
            )
        if param_name.endswith(".in"):
            temp_value = []
            for v in value:
                if v.startswith("[") and v.endswith("]") and "," in v:
                    v = v[1:-1].split(",")
                    temp_value.extend(v)
                else:
                    temp_value.append(v)
            value = temp_value
        params_map[param_name] = value
    logger.debug("Extracted object params: %s", params_map)
    return None if raw else params_map


def handle_encodings(request: Request) -> QueryParams:
    """Handle encodings.

    WSGIRequest does not support Unicode values in the query string.
    WSGIRequest handling has a history of drifting behavior between
    combinations of Python versions, Django versions and DRF versions.
    Django changed its QUERY_STRING handling here:
    https://goo.gl/WThXo6. DRF 3.4.7 changed its behavior here:
    https://goo.gl/0ojIIO.

    Args:
        request: Request object.

    Returns:
        QueryParams: Query parameters.
    """
    try:
        return QueryParams(request.GET)
    except UnicodeEncodeError:
        pass

    s = request.environ.get("QUERY_STRING", "")
    try:
        s = s.encode("utf-8")
    except UnicodeDecodeError:
        pass
    return QueryParams(s)


class QueryParams(QueryDict):
    """Extension of Django's QueryDict.

    Instantiated from a DRF Request object, and returns
     a mutable QueryDict subclass. Also adds methods that
     might be useful for our use-case.
    """

    def __init__(self, query_params, *args, **kwargs):
        """Initialize the QueryParams object."""
        if hasattr(query_params, "urlencode"):
            query_string = query_params.urlencode()
        else:
            assert isinstance(query_params, (str, bytes))
            query_string = query_params
        kwargs["mutable"] = True
        super().__init__(query_string, *args, **kwargs)

    def add(self, key, value):
        """Add a key/value pair to the QueryDict.

        Method to accept a list of values and append to flat list.
        QueryDict.appendlist(), if given a list, will append the list,
        which creates nested lists. In most cases, we want to be able
        to pass in a list (for convenience) but have it appended into
        a flattened list.
        TODO: Possibly throw an error if add() is used on a non-list param.
        """
        if isinstance(value, list):
            for val in value:
                self.appendlist(key, val)
        else:
            self.appendlist(key, value)


@runtime_checkable
class HasRequestProperty(Protocol):
    """Protocol for request property."""

    request: Request


@runtime_checkable
class HasInitializeRequest(Protocol):
    """Protocol for initialize_request method."""

    # pylint: disable=W0246
    def initialize_request(self, request: Request, *args, **kwargs) -> Request:
        """Initialize the request object."""
        return super().initialize_request(request, *args, **kwargs)


@runtime_checkable
class HasGetRenderers(Protocol):
    """Protocol for get_renderers method."""

    # pylint: disable=W0246
    def get_renderers(self) -> list[BaseRenderer]:
        """Get renderers."""
        return super().get_renderers()


class WithDynamicViewSetMixin(
    HasInitializeRequest, HasGetRenderers, HasRequestProperty
):
    """A ViewSet that can support dynamic API features.

    Attributes:
      features: A list of features supported by the ViewSet.
      meta: Extra data that is added to the response by the DynamicRenderer.
    """

    DEBUG = "debug"
    SIDELOADING = "sideloading"
    PATCH_ALL = "patch-all"
    INCLUDE = "include[]"
    EXCLUDE = "exclude[]"
    FILTER = "filter{}"
    SORT = "sort[]"
    PAGE = settings.PAGE_QUERY_PARAM
    PER_PAGE = settings.PAGE_SIZE_QUERY_PARAM

    # TODO: add support for `sort{}`
    pagination_class = DynamicPageNumberPagination
    metadata_class = DynamicMetadata
    features = (
        DEBUG,
        INCLUDE,
        EXCLUDE,
        FILTER,
        PAGE,
        PER_PAGE,
        SORT,
        SIDELOADING,
        PATCH_ALL,
    )
    meta = None
    filter_backends = (DynamicFilterBackend, DynamicSortingFilter)

    def initialize_request(self, request: Request, *args, **kwargs) -> Request:
        """Initialize the request object.

        Override DRF initialize_request() method to swap request.GET
        (which is aliased by request.query_params) with a mutable instance
        of QueryParams, and to convert request MergeDict to a subclass of dict
        for consistency (MergeDict is not a subclass of dict)

        Args:
            request: Request object.
            *args: Args.
            **kwargs: Kwargs.

        Returns:
            Request: Request object.
        """
        request.GET = handle_encodings(request)
        return super().initialize_request(request, *args, **kwargs)

    def get_renderers(self) -> list[BaseRenderer]:
        """Optionally block Browsable API rendering.

        Returns:
            list[BaseRenderer]: List of renderers.
        """
        renderers = super().get_renderers()
        if settings.ENABLE_BROWSABLE_API is False:
            return [r for r in renderers if not isinstance(r, BrowsableAPIRenderer)]
        return renderers

    def get_request_feature(self, name, raw: bool = False):
        """Parses the request for a particular feature.

        Arguments:
          name: A feature name.
          raw: bool -

        Returns:
          A feature parsed from the URL if the feature is supported, or None.
        """
        name_is_feature = name in self.features
        request = self.request
        if "[]" in name:
            logger.debug("Using array-type feature: %s", name)
            return request.query_params.getlist(name) if name_is_feature else None
        elif "{}" in name:
            logger.debug(
                "Using object-type feature (keys are not consistent): %s", name
            )
            return (
                _extract_object_params(request, name, raw=raw)
                if name_is_feature
                else {}
            )
        logger.debug("Using single-type feature: %s", name)
        return request.query_params.get(name) if name_is_feature else None

    def get_queryset(self, queryset=None):  # pylint: disable=unused-argument
        """Returns a queryset for this request.

        Arguments:
          queryset: Optional root-level queryset.
        """
        serializer = self.get_serializer()
        return getattr(self, "queryset", serializer.Meta.model.objects.all())

    def get_request_fields(self):
        """Parses the INCLUDE and EXCLUDE features.

        Extracts the dynamic field features from the request parameters
        into a field map that can be passed to a serializer.

        Returns:
          A nested dict mapping serializer keys to
          True (include) or False (exclude).
        """
        if hasattr(self, "_request_fields"):
            return self._request_fields

        include_fields = self.get_request_feature(self.INCLUDE)
        exclude_fields = self.get_request_feature(self.EXCLUDE)
        request_fields = {}
        for fields, include in ((include_fields, True), (exclude_fields, False)):
            if fields is None:
                continue
            for field in fields:
                field_segments = field.split(".")
                num_segments = len(field_segments)
                current_fields = request_fields
                for i, segment in enumerate(field_segments):
                    last = i == num_segments - 1
                    if segment:
                        if last:
                            current_fields[segment] = include
                        else:
                            if segment not in current_fields:
                                current_fields[segment] = {}
                            current_fields = current_fields[segment]
                    elif not last:
                        # empty segment must be the last segment
                        raise exceptions.ParseError(f'"{field}" is not a valid field.')

        self._request_fields = request_fields
        return request_fields

    def get_request_patch_all(self):
        """Get request patch-all value."""
        patch_all = self.get_request_feature(self.PATCH_ALL)
        if not patch_all:
            return None
        patch_all = patch_all.lower()
        if patch_all == "query":
            pass
        elif is_truthy(patch_all):
            patch_all = True
        else:
            raise exceptions.ParseError(
                f'"{patch_all}" is not valid for {self.PATCH_ALL}'
            )
        return patch_all

    def get_request_debug(self):
        """Get request debug value."""
        debug = self.get_request_feature(self.DEBUG)
        return is_truthy(debug) if debug is not None else None

    def get_request_sideloading(self):
        """Get request sideloading value."""
        sideloading = self.get_request_feature(self.SIDELOADING)
        return is_truthy(sideloading) if sideloading is not None else None

    def is_update(self):
        """Return True if the request is an update request."""
        return self.request and self.request.method.upper() in UPDATE_REQUEST_METHODS

    def is_delete(self):
        """Return True if the request is a delete request."""
        return self.request and self.request.method.upper() == DELETE_REQUEST_METHOD

    def get_serializer(self, *args, **kwargs):
        """Return a serializer instance."""
        if "request_fields" not in kwargs:
            kwargs["request_fields"] = self.get_request_fields()
        if "sideloading" not in kwargs:
            kwargs["sideloading"] = self.get_request_sideloading()
        if "debug" not in kwargs:
            kwargs["debug"] = self.get_request_debug()
        if "envelope" not in kwargs:
            kwargs["envelope"] = True
        if self.is_update():
            kwargs["include_fields"] = "*"
        return super().get_serializer(*args, **kwargs)

    def paginate_queryset(self, *args, **kwargs):
        """Paginate the queryset if pagination is enabled."""
        if self.PAGE not in self.features:
            return
        query_params = self.request.query_params
        per_page = self.PER_PAGE
        # make sure pagination is enabled
        if per_page not in self.features and per_page in query_params:
            # remove per_page if it is disabled
            query_params[per_page] = None
        return super().paginate_queryset(*args, **kwargs)

    def _prefix_inex_params(self, request, feature, prefix):
        """Prefix include/exclude params with field_name."""
        values = self.get_request_feature(feature)
        if not values:
            return
        del request.query_params[feature]
        request.query_params.add(feature, [prefix + val for val in values])

    def list_related(self, request, pk=None, field_name=None):
        """List related.

        Fetch related object(s), as if side-loaded (used to support
        link objects).

        This method gets mapped to `/<resource>/<pk>/<field_name>/` by
        DynamicRouter for all DynamicRelationField fields. Generally,
        this method probably shouldn't be overridden.

        An alternative implementation would be to generate reverse queries.
        For an exploration of that approach, see:
            https://gist.github.com/ryochiji/54687d675978c7d96503
        """
        # Explicitly disable support filtering. Applying filters to this
        # endpoint would require us to pass through side-load filters, which
        # can have unintended consequences when applied asynchronously.
        if self.get_request_feature(self.FILTER):
            raise ValidationError("Filtering is not enabled on relation endpoints.")

        # Prefix include/exclude filters with field_name, so it's scoped to
        # the parent object.
        field_prefix = field_name + "."
        self._prefix_inex_params(request, self.INCLUDE, field_prefix)
        self._prefix_inex_params(request, self.EXCLUDE, field_prefix)

        # Filter for parent object, include related field.
        self.request.query_params.add("filter{pk}", pk)
        self.request.query_params.add(self.INCLUDE, field_prefix)

        # Get serializer and field.
        serializer = self.get_serializer()
        field = serializer.fields.get(field_name)
        if field is None:
            raise ValidationError(f'Unknown field: "{field_name}".')

        # Query for root object, with related field prefetched
        queryset = self.get_queryset()
        queryset = self.filter_queryset(queryset)
        obj = queryset.first()

        if not obj:
            return Response("Not found", status=404)

        # Serialize the related data. Use the field's serializer to ensure
        # it's configured identically to the sideload case. One difference
        # is we need to set `envelope=True` to get the sideload-processor
        # applied.
        related_szr = field.get_serializer(envelope=True)
        try:
            # TODO(ryo): Probably should use field.get_attribute() but that
            #            seems to break a bunch of things. Investigate later.
            related_szr.instance = getattr(obj, field.source)
        except ObjectDoesNotExist:
            # See:
            # http://jsonapi.org/format/#fetching-relationships-responses-404
            # This is a case where the "link URL exists but the relationship
            # is empty" and therefore must return a 200.
            return Response({}, status=200)

        return Response(related_szr.data)

    def get_extra_filters(self, request):  # pylint: disable=unused-argument
        """Get extra filters.

        Override this method to enable addition of extra filters
         (i.e., a Q()) so custom filters can be added to the queryset without
         running into https://code.djangoproject.com/ticket/18437
         which, without this, would mean that filters added to the queryset
         after this is called may not behave as expected.
        """
        return None


class DynamicModelViewSet(WithDynamicViewSetMixin, viewsets.ModelViewSet):
    """A ModelViewSet that supports dynamic API features."""

    ENABLE_BULK_PARTIAL_CREATION = settings.ENABLE_BULK_PARTIAL_CREATION
    ENABLE_BULK_UPDATE = settings.ENABLE_BULK_UPDATE
    ENABLE_PATCH_ALL = settings.ENABLE_PATCH_ALL

    def _get_bulk_payload(self, request):
        """Get bulk payload from request."""
        plural_name = self.get_serializer_class().get_plural_name()
        data = request.data
        if isinstance(data, list):
            return data
        elif plural_name in data and len(data) == 1:
            return data[plural_name]
        return None

    def _bulk_update(self, data, partial=False):
        """Bulk update records."""
        # Restrict the update to the filtered queryset.
        serializer = self.get_serializer(
            self.filter_queryset(self.get_queryset()),
            data=data,
            many=True,
            partial=partial,
        )
        serializer.is_valid(raise_exception=True)
        self.perform_update(serializer)
        return Response(serializer.data, status=status.HTTP_200_OK)

    def _validate_patch_all(self, data):
        """Validate patch-all data."""
        if not isinstance(data, dict):
            raise ValidationError("Patch-all data must be in object form")
        serializer = self.get_serializer()
        fields = serializer.get_all_fields()
        validated = {}
        for name, value in data.items():
            field = fields.get(name, None)
            if field is None:
                raise ValidationError(f'Unknown field: "{name}"')
            source = field.source or name
            if source == "*" or field.read_only:
                raise ValidationError(f'Cannot update field: "{name}"')
            validated[source] = value
        return validated

    def _patch_all_query(self, queryset, data):
        """Update by queryset."""
        # update by queryset
        try:
            return queryset.update(**data)
        except Exception as e:
            raise ValidationError(
                "Failed to bulk-update records:\n" f"{str(e)}\n" f"Data: {str(data)}"
            ) from e

    def _patch_all_loop(self, queryset, data):
        """Update by transaction loop."""
        # update by transaction loop
        updated = 0
        try:
            with transaction.atomic():
                for record in queryset:
                    for k, v in data.items():
                        setattr(record, k, v)
                    record.save()
                    updated += 1
                return updated
        except IntegrityError as e:
            raise ValidationError(
                "Failed to update records:\n" f"{str(e)}\n" f"Data: {str(data)}"
            ) from e

    def _patch_all(self, data, query=False):
        """Update all records in a queryset."""
        queryset = self.filter_queryset(self.get_queryset())
        data = self._validate_patch_all(data)
        updated = (
            self._patch_all_query(queryset, data)
            if query
            else self._patch_all_loop(queryset, data)
        )
        return Response({"meta": {"updated": updated}}, status=status.HTTP_200_OK)

    def update(self, request, *args, **kwargs):
        """Update one or more model instances.

        If ENABLE_BULK_UPDATE is set, multiple previously-fetched records
        may be updated in a single call, provided their IDs.

        If ENABLE_PATCH_ALL is set, multiple records
        may be updated in a single PATCH call, even without knowing their IDs.

        *WARNING*: ENABLE_PATCH_ALL should be considered an advanced feature
        and used with caution. This feature must be enabled at the viewset level
        and must also be requested explicitly by the client
        via the "patch-all" query parameter.

        This parameter can have one of the following values:

            true (or 1): records will be fetched and then updated in a
             transaction loop
              - The `Model.save` method will be called and model signals
                will run
              - This can be slow if there are too many signals
               or many records in the query.
              - This is considered the more safe and default behavior
            query: records will be updated in a single query
              - The `QuerySet.update` method will be called and model
               signals will not run
              - This will be fast, but may break data constraints that
               are controlled by signals
              - This is considered unsafe but useful in certain situations

        The server's successful response to a patch-all request
        will NOT include any individual records.
        Instead, the response content will containa "meta" object
         with an "updated" count of updated records.

        Examples:

        Update one dog:

            PATCH /dogs/1/
            {
                'fur': 'white'
            }

        Update many dogs by ID:

            PATCH /dogs/
            [
                {'id': 1, 'fur': 'white'},
                {'id': 2, 'fur': 'black'},
                {'id': 3, 'fur': 'yellow'}
            ]

        Update all dogs in a query:

            PATCH /dogs/?filter{fur.contains}=brown&patch-all=true
            {
                'fur': 'gold'
            }
        """  # noqa
        if self.ENABLE_BULK_UPDATE:
            patch_all = self.get_request_patch_all()
            if self.ENABLE_PATCH_ALL and patch_all:
                # patch-all update
                data = request.data
                return self._patch_all(data, query=patch_all == "query")
            else:
                # bulk payload update
                partial = "partial" in kwargs
                bulk_payload = self._get_bulk_payload(request)
                if bulk_payload:
                    return self._bulk_update(bulk_payload, partial)

        # singular update
        try:
            return super().update(request, *args, **kwargs)
        except AssertionError as e:
            err = str(e)
            if "Fix your URL conf" in err:
                # this error is returned by DRF if a client
                # makes an update request (PUT or PATCH) without an ID
                # since DREST supports bulk updates with IDs contained in data,
                # we return a 400 instead of a 500 for this case,
                # as this is not considered a misconfiguration
                raise exceptions.ValidationError(err)
            else:
                raise

    def _create_many(self, data):
        """Create many model instances in bulk."""
        items = []
        errors = []
        result = {}
        serializers = []

        for entry in data:
            serializer = self.get_serializer(data=entry)
            try:
                serializer.is_valid(raise_exception=True)
            except exceptions.ValidationError as e:
                errors.append({"detail": str(e), "source": entry})
            else:
                if self.ENABLE_BULK_PARTIAL_CREATION:
                    self.perform_create(serializer)
                    items.append(serializer.to_representation(serializer.instance))
                else:
                    serializers.append(serializer)
        if not self.ENABLE_BULK_PARTIAL_CREATION and not errors:
            for serializer in serializers:
                self.perform_create(serializer)
                items.append(serializer.to_representation(serializer.instance))

        # Populate serialized data to the result.
        result = SideloadingProcessor(self.get_serializer(), items).data

        # Include errors if any.
        if errors:
            result["errors"] = errors

        code = status.HTTP_201_CREATED if not errors else status.HTTP_400_BAD_REQUEST

        return Response(result, status=code)

    def create(self, request, *args, **kwargs):
        """Create one or more model instances.

        Either create a single or many model instances in bulk
        using the Serializer's many=True ability from Django REST >= 2.2.5.

        The data can be represented by the serializer name (single or plural
        forms), dict or list.

        Examples:
        POST /dogs/
        {
          "name": "Fido",
          "age": 2
        }

        POST /dogs/
        {
          "dog": {
            "name": "Lucky",
            "age": 3
          }
        }

        POST /dogs/
        {
          "dogs": [
            {"name": "Fido", "age": 2},
            {"name": "Lucky", "age": 3}
          ]
        }

        POST /dogs/
        [
            {"name": "Fido", "age": 2},
            {"name": "Lucky", "age": 3}
        ]
        """
        bulk_payload = self._get_bulk_payload(request)
        if bulk_payload:
            return self._create_many(bulk_payload)
        return super().create(request, *args, **kwargs)

    def _destroy_many(self, data):
        """Destroy many model instances in bulk."""
        instances = (
            self.get_queryset().filter(id__in=[d["id"] for d in data]).distinct()
        )
        for instance in instances:
            self.check_object_permissions(self.request, instance)
            self.perform_destroy(instance)
        return Response(status=status.HTTP_204_NO_CONTENT)

    def destroy(self, request, *args, **kwargs):
        """Either delete a single or many model instances in bulk.

        DELETE /dogs/
        {
            "dogs": [
                {"id": 1},
                {"id": 2}
            ]
        }

        DELETE /dogs/
        [
            {"id": 1},
            {"id": 2}
        ]
        """
        bulk_payload = self._get_bulk_payload(request)
        if bulk_payload:
            return self._destroy_many(bulk_payload)
        lookup_url_kwarg = self.lookup_url_kwarg or self.lookup_field
        if lookup_url_kwarg not in kwargs:
            # assume that it is a poorly formatted bulk request
            return Response(status=status.HTTP_405_METHOD_NOT_ALLOWED)
        return super().destroy(request, *args, **kwargs)



"""This module contains base classes for DREST."""

from dynamic_rest.utils import model_from_definition


class DynamicSerializerBase(object):
    """Base class for all DREST serializers."""

    pass


def resettable_cached_property(func):
    """Decorator to add cached computed properties to an object.

    Similar to Django's `cached_property` decorator, except stores
    all the data under a single well-known key so that it can easily
    be blown away.
    """

    def wrapper(self):
        self._resettable_cached_properties = (  # pylint: disable=protected-access
            cache
        ) = getattr(self, "_resettable_cached_properties", {})
        func_name = func.__name__
        if func_name not in cache:
            cache[func_name] = func(self)
        return cache[func_name]

    # Returns a property whose getter is the 'wrapper' function
    return property(wrapper)


def cacheable_object(cls):
    """Decorator to add a reset() method that clears cached data.

    Cached data is set by the @resettable_cached_property decorator.
    Technically this could be a mixin...
    """

    def reset(self):
        """Reset the cached properties."""
        self._resettable_cached_properties = {}  # pylint: disable=protected-access

    cls.reset = reset
    return cls


@cacheable_object
class CacheableFieldMixin(object):
    """
    Cachable field mixin.

    Override Field.root and Field.context to make fields/serializers
    cacheable and reusable.

    The DRF version uses @cached_property which doesn't have a
    public API for resetting.
    This version uses normal object variables with and adds a `reset()` API.
    """

    @resettable_cached_property
    def root(self):
        """Find the top level root."""
        root = self
        while root.parent is not None:
            root = root.parent
        return root

    @resettable_cached_property
    def context(self):
        """Get the context from the root."""
        return getattr(self.root, "_context", {})


class GetModelMixin(object):
    """
    Mixin to retrieve model hashid.

    Implementation from
    https://github.com/evenicoulddoit/django-rest-framework-serializer-extensions
    """

    def __init__(self, *args, **kwargs):
        """Initialise the GetModelMixin."""
        self.model = kwargs.pop("model", None)
        super().__init__(*args, **kwargs)

    def get_model(self):
        """
        Return the model to generate the HashId for.

        By default, this will equal the model defined within the Meta of the
        ModelSerializer, but can be redefined either during initialisation
        of the Field, or by providing a get_<field_name>_model method on the
        parent serializer.

        The Meta can either explicitly define a model, or provide a
        dot-delimited string path to it.
        """
        model = self.model
        if model is None:
            if model is None:
                custom_fn_name = f"get_{self.field_name}_model"
                parent = self.parent
                if hasattr(parent, custom_fn_name):
                    self.model = getattr(parent, custom_fn_name)()
                else:
                    try:
                        self.model = parent.Meta.model
                    except AttributeError as exc:
                        raise AssertionError(
                            f'No "model" value passed to field "{type(self).__name__}"'
                        ) from exc
            elif isinstance(model, str):
                self.model = model_from_definition(model)
            else:
                self.model = model
        return self.model



"""Tests for viewsets."""
import json
import os

from django.test.client import RequestFactory
from rest_framework import exceptions, status
from rest_framework.request import Request

from dynamic_rest.datastructures import FilterNode
from dynamic_rest.filters import DynamicFilterBackend
from dynamic_rest.filters.base import _get_requested_filters
from tests.models import Dog, Group, User
from tests.serializers import GroupSerializer
from tests.setup import create_fixture
from tests.viewsets import GroupNoMergeDictViewSet, UserViewSet

if os.getenv("DATABASE_URL"):
    from tests.test_cases import ResetTestCase as TestCase
else:
    from tests.test_cases import TestCase


class TestUserViewSet(TestCase):
    """Test case for UserViewSet."""

    def setUp(self):
        """Set up test case."""
        self.view = UserViewSet()
        self.rf = RequestFactory()

    def test_get_request_fields(self):
        """Test get request fields."""
        request = Request(
            self.rf.get(
                "/users/",
                {
                    "include[]": ["name", "groups.permissions"],
                    "exclude[]": ["groups.name"],
                },
            )
        )
        self.view.request = request
        fields = self.view.get_request_fields()

        self.assertEqual(
            {"groups": {"name": False, "permissions": True}, "name": True}, fields
        )

    def test_get_request_fields_disabled(self):
        """Test get request fields disabled."""
        self.view.features = self.view.INCLUDE
        request = Request(
            self.rf.get(
                "/users/",
                {"include[]": ["name", "groups"], "exclude[]": ["groups.name"]},
            )
        )
        self.view.request = request
        fields = self.view.get_request_fields()

        self.assertEqual({"groups": True, "name": True}, fields)

    def test_get_request_fields_invalid(self):
        """Test get request fields invalid."""
        for invalid_field in ("groups..name", "groups.."):
            request = Request(self.rf.get("/users/", {"include[]": [invalid_field]}))
            self.view.request = request
            self.assertRaises(exceptions.ParseError, self.view.get_request_fields)

    def test_filter_extraction(self):
        """Test filter extraction."""
        filters_map = {
            "attr": ["bar"],
            "attr2.eq": ["bar"],
            "-attr3": ["bar"],
            "rel|attr1": ["val"],
            "-rel|attr2": ["val"],
            "rel.attr": ["baz"],
            "rel.bar|attr": ["val"],
            "attr4.lt": ["val"],
            "attr5.in": ["val1", "val2", "val3"],
        }

        backend = DynamicFilterBackend()
        # pylint: disable=protected-access
        out = _get_requested_filters(
            getattr(backend, "view", None), filters_map=filters_map
        )
        self.assertEqual(out["_include"]["attr"].value, "bar")
        self.assertEqual(out["_include"]["attr2"].value, "bar")
        self.assertEqual(out["_exclude"]["attr3"].value, "bar")
        self.assertEqual(out["rel"]["_include"]["attr1"].value, "val")
        self.assertEqual(out["rel"]["_exclude"]["attr2"].value, "val")
        self.assertEqual(out["_include"]["rel__attr"].value, "baz")
        self.assertEqual(out["rel"]["bar"]["_include"]["attr"].value, "val")
        self.assertEqual(out["_include"]["attr4__lt"].value, "val")
        self.assertEqual(len(out["_include"]["attr5__in"].value), 3)

    def test_is_null_casting(self):
        """Test is null casting."""
        filters_map = {
            "f1.isnull": [True],
            "f2.isnull": [["a"]],
            "f3.isnull": ["true"],
            "f4.isnull": ["1"],
            "f5.isnull": [1],
            "f6.isnull": [False],
            "f7.isnull": [[]],
            "f8.isnull": ["false"],
            "f9.isnull": ["0"],
            "f10.isnull": [0],
            "f11.isnull": [""],
            "f12.isnull": [None],
        }

        backend = DynamicFilterBackend()
        # pylint: disable=protected-access
        out = _get_requested_filters(
            getattr(backend, "view", None), filters_map=filters_map
        )

        self.assertEqual(out["_include"]["f1__isnull"].value, True)
        self.assertEqual(out["_include"]["f2__isnull"].value, ["a"])
        self.assertEqual(out["_include"]["f3__isnull"].value, True)
        self.assertEqual(out["_include"]["f4__isnull"].value, True)
        self.assertEqual(out["_include"]["f5__isnull"].value, 1)

        self.assertEqual(out["_include"]["f6__isnull"].value, False)
        self.assertEqual(out["_include"]["f7__isnull"].value, [])
        self.assertEqual(out["_include"]["f8__isnull"].value, False)
        self.assertEqual(out["_include"]["f9__isnull"].value, False)
        self.assertEqual(out["_include"]["f10__isnull"].value, False)
        self.assertEqual(out["_include"]["f11__isnull"].value, False)
        self.assertEqual(out["_include"]["f12__isnull"].value, None)

    def test_nested_filter_rewrite(self):
        """Test nested filter rewrite."""
        node = FilterNode(["members", "id"], "in", [1])
        gs = GroupSerializer(include_fields="*")
        filter_key, _ = node.generate_query_key(gs)
        self.assertEqual(filter_key, "users__id__in")


class TestMergeDictConvertsToDict(TestCase):
    """Test case for MergeDict behavior in DRF 3.2."""

    def setUp(self):
        """Set up test case."""
        self.fixture = create_fixture()
        self.view = GroupNoMergeDictViewSet.as_view({"post": "create"})
        self.rf = RequestFactory()

    def test_merge_dict_request(self):
        """Test merge dict request."""
        data = {"name": "miao", "random_input": [1, 2, 3]}
        # Django test submits data as multipart-form by default,
        # which results in request.data being a MergeDict.
        # Wrote UserNoMergeDictViewSet to raise an exception (return 400)
        # if request.data ends up as MergeDict, is not a dict, or
        # is a dict of lists.
        request = self.rf.post("/groups/", data)
        try:
            response = self.view(request)
            self.assertEqual(response.status_code, 201)
        except NotImplementedError as e:
            message = f"{str(e)}"
            if "request.FILES" not in message:
                self.fail(f"Unexpected error: {message}")
            # otherwise, this is a known DRF 3.2 bug


class BulkUpdateTestCase(TestCase):
    """Test case for bulk update."""

    def setUp(self):
        """Set up test case."""
        self.fixture = create_fixture()

    def test_bulk_update_default_style(self):
        """Test that PATCH request partially updates all submitted resources."""
        data = [{"id": 1, "fur": "grey"}, {"id": 2, "fur": "grey"}]
        response = self.client.patch(
            "/dogs/", json.dumps(data), content_type="application/json"
        )
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertTrue("dogs" in response.data)
        self.assertTrue(len(response.data["dogs"]) == 2)
        self.assertTrue(
            all(Dog.objects.get(id=pk).fur_color == "grey" for pk in (1, 2))
        )

    def test_bulk_update_drest_style(self):
        """Test that PATCH request partially updates all submitted resources."""
        # test to make sure both string '2' and integer 1 resolve correctly
        data = {"dogs": [{"id": 1, "fur": "grey"}, {"id": "2", "fur": "grey"}]}
        response = self.client.patch(
            "/dogs/", json.dumps(data), content_type="application/json"
        )
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertTrue("dogs" in response.data)

    def test_bulk_update_with_filter(self):
        """Test that you can patch inside the filtered queryset."""
        data = [{"id": 3, "fur": "gold"}]
        response = self.client.patch(
            "/dogs/?filter{fur.contains}=brown",
            json.dumps(data),
            content_type="application/json",
        )
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertTrue(Dog.objects.get(id=3).fur_color == "gold")

    def test_bulk_update_fail_without_query_param(self):
        """Test bulk update fail without query param.

        Test that an update-all PATCH request will fail
        if not explicitly using update-all syntax
        """
        for data in [{"fur": "grey"}], []:
            response = self.client.patch(
                "/dogs/?filter{fur.contains}=brown",
                json.dumps(data),
                content_type="application/json",
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

    def test_patch_all_validation(self):
        """Test patch all validation."""
        # wrong format
        data = [{"fur": "grey"}]
        response = self.client.patch(
            "/dogs/?patch-all=true", json.dumps(data), content_type="application/json"
        )
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

        # wrong field
        data = {"fury": "grey"}
        response = self.client.patch(
            "/dogs/?patch-all=true", json.dumps(data), content_type="application/json"
        )
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertTrue("fury" in response.content.decode("utf-8"))

        # non-source field
        data = {"is_red": True, "fur": "red"}
        response = self.client.patch(
            "/dogs/?patch-all=true", json.dumps(data), content_type="application/json"
        )
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertTrue("is_red" in response.content.decode("utf-8"))

    def test_patch_all(self):
        """Test patch all."""
        # the correct format for a patch-all request
        data = {"fur": "grey"}
        response = self.client.patch(
            "/dogs/?patch-all=true", json.dumps(data), content_type="application/json"
        )
        self.assertEqual(response.status_code, status.HTTP_200_OK, response.content)
        content = json.loads(response.content.decode("utf-8"))
        num_dogs = Dog.objects.all().count()
        self.assertEqual(num_dogs, content["meta"]["updated"])
        self.assertEqual(
            num_dogs,
            Dog.objects.filter(fur_color="grey").count(),
        )


class BulkCreationTestCase(TestCase):
    """Test case for bulk creation."""

    def test_post_single(self):
        """Test post single.

        Test that POST request with single resource only creates a single
        resource.
        """
        data = {"name": "foo", "random_input": [1, 2, 3]}
        response = self.client.post(
            "/groups/", json.dumps(data), content_type="application/json"
        )
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        self.assertEqual(1, Group.objects.all().count())

    def test_post_bulk_from_resource_plural_name(self):
        """Test post bulk from resource plural name."""
        data = {
            "groups": [
                {"name": "foo", "random_input": [1, 2, 3]},
                {"name": "bar", "random_input": [4, 5, 6]},
            ]
        }
        response = self.client.post(
            "/groups/", json.dumps(data), content_type="application/json"
        )
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        self.assertEqual(2, Group.objects.all().count())

    def test_post_bulk_from_list(self):
        """Test post bulk from list.

        Test that POST request with multiple resources created all posted
        resources.
        """
        data = [
            {
                "name": "foo",
                "random_input": [1, 2, 3],
            },
            {
                "name": "bar",
                "random_input": [4, 5, 6],
            },
        ]
        response = self.client.post(
            "/groups/", json.dumps(data), content_type="application/json"
        )
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        self.assertEqual(2, Group.objects.all().count())
        self.assertEqual(
            ["foo", "bar"], list(Group.objects.all().values_list("name", flat=True))
        )

    def test_post_bulk_with_existing_items_and_disabled_partial_creation(self):
        """Test post bulk with existing items and disabled partial creation."""
        data = [{"name": "foo"}, {"name": "bar"}]
        Group.objects.create(name="foo")
        response = self.client.post(
            "/groups/", json.dumps(data), content_type="application/json"
        )
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertEqual(1, Group.objects.all().count())
        self.assertTrue("errors" in response.data)

    def test_post_bulk_with_sideloaded_results(self):
        """Test post bulk with sideloaded results."""
        u1 = User.objects.create(name="foo", last_name="bar")
        u2 = User.objects.create(name="foo", last_name="baz")
        data = [
            {
                "name": "foo",
                "members": [u1.pk],
            },
            {
                "name": "bar",
                "members": [u2.pk],
            },
        ]
        response = self.client.post(
            "/groups/?include[]=members.",
            json.dumps(data),
            content_type="application/json",
        )
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        resp_data = response.data

        # Check top-level keys
        self.assertEqual({"users", "groups"}, set(resp_data.keys()))

        # Should be 2 of each
        self.assertEqual(2, len(resp_data["users"]))
        self.assertEqual(2, len(resp_data["groups"]))


class BulkDeletionTestCase(TestCase):
    """Test case for bulk deletion."""

    def setUp(self):
        """Set up test case."""
        self.fixture = create_fixture()
        self.ids = [i.pk for i in self.fixture.dogs]
        self.ids_to_delete = self.ids[:2]

    def test_bulk_delete_default_style(self):
        """Test bulk delete default style."""
        data = [{"id": i} for i in self.ids_to_delete]
        response = self.client.delete(
            "/dogs/",
            json.dumps(data),
            content_type="application/json",
        )
        self.assertEqual(response.status_code, status.HTTP_204_NO_CONTENT)
        self.assertEqual(Dog.objects.filter(id__in=self.ids_to_delete).count(), 0)

    def test_bulk_delete_drest_style(self):
        """Test bulk delete drest style."""
        data = {"dogs": [{"id": i} for i in self.ids_to_delete]}
        response = self.client.delete(
            "/dogs/",
            json.dumps(data),
            content_type="application/json",
        )
        self.assertEqual(response.status_code, status.HTTP_204_NO_CONTENT)
        self.assertEqual(Dog.objects.filter(id__in=self.ids_to_delete).count(), 0)

    def test_bulk_delete_single(self):
        """Test bulk delete single."""
        response = self.client.delete(f"/dogs/{self.ids_to_delete[0]}")
        self.assertEqual(response.status_code, status.HTTP_204_NO_CONTENT)

    def test_bulk_delete_invalid_single(self):
        """Test bulk delete invalid single."""
        data = {"dog": {"id": self.ids_to_delete[0]}}
        response = self.client.delete(
            "/dogs/",
            json.dumps(data),
            content_type="application/json",
        )
        self.assertEqual(response.status_code, status.HTTP_405_METHOD_NOT_ALLOWED)

    def test_bulk_delete_invalid(self):
        """Test bulk delete invalid."""
        data = {"id": self.ids_to_delete[0]}
        response = self.client.delete(
            "/dogs/",
            json.dumps(data),
            content_type="application/json",
        )
        self.assertEqual(response.status_code, status.HTTP_405_METHOD_NOT_ALLOWED)

    def test_delete_on_nonexistent_raises_404(self):
        """Test delete on nonexistent raises 404."""
        response = self.client.delete("/dogs/31415")
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)



# -*- coding: utf-8 -*-
# Generated by Django 1.9 on 2016-08-08 13:28
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("tests", "0003_auto_20160401_1656"),
    ]

    operations = [
        migrations.AddField(
            model_name="user",
            name="is_dead",
            field=models.NullBooleanField(default=False),
        ),
    ]



# -*- coding: utf-8 -*-
# flake8: noqa
# Generated by Django 1.9 on 2016-01-27 16:29
from __future__ import unicode_literals

import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):
    initial = True

    dependencies = []

    operations = [
        migrations.CreateModel(
            name="A",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("name", models.TextField(blank=True)),
            ],
        ),
        migrations.CreateModel(
            name="B",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "a",
                    models.OneToOneField(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="b",
                        to="tests.A",
                    ),
                ),
            ],
        ),
        migrations.CreateModel(
            name="C",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "b",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="cs",
                        to="tests.B",
                    ),
                ),
            ],
        ),
        migrations.CreateModel(
            name="Cat",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("name", models.TextField()),
            ],
        ),
        migrations.CreateModel(
            name="D",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("name", models.TextField(blank=True)),
            ],
        ),
        migrations.CreateModel(
            name="Dog",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("name", models.TextField()),
                ("fur_color", models.TextField()),
                ("origin", models.TextField()),
            ],
        ),
        migrations.CreateModel(
            name="Event",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("name", models.TextField()),
                ("status", models.TextField(default=b"current")),
            ],
        ),
        migrations.CreateModel(
            name="Group",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("name", models.TextField()),
            ],
        ),
        migrations.CreateModel(
            name="Horse",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("name", models.TextField()),
                ("origin", models.TextField()),
            ],
        ),
        migrations.CreateModel(
            name="Location",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("name", models.TextField()),
                ("blob", models.TextField()),
            ],
        ),
        migrations.CreateModel(
            name="Permission",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("name", models.TextField()),
                ("code", models.IntegerField()),
            ],
        ),
        migrations.CreateModel(
            name="Profile",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("display_name", models.TextField()),
                ("thumbnail_url", models.TextField(blank=True, null=True)),
            ],
        ),
        migrations.CreateModel(
            name="User",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("name", models.TextField()),
                ("last_name", models.TextField()),
                (
                    "groups",
                    models.ManyToManyField(related_name="users", to="tests.Group"),
                ),
                (
                    "location",
                    models.ForeignKey(
                        blank=True,
                        null=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        to="tests.Location",
                    ),
                ),
                (
                    "permissions",
                    models.ManyToManyField(related_name="users", to="tests.Permission"),
                ),
            ],
        ),
        migrations.CreateModel(
            name="Zebra",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("name", models.TextField()),
                ("origin", models.TextField()),
            ],
        ),
        migrations.AddField(
            model_name="profile",
            name="user",
            field=models.OneToOneField(
                on_delete=django.db.models.deletion.CASCADE, to="tests.User"
            ),
        ),
        migrations.AddField(
            model_name="group",
            name="permissions",
            field=models.ManyToManyField(related_name="groups", to="tests.Permission"),
        ),
        migrations.AddField(
            model_name="event",
            name="location",
            field=models.ForeignKey(
                blank=True,
                null=True,
                on_delete=django.db.models.deletion.CASCADE,
                to="tests.Location",
            ),
        ),
        migrations.AddField(
            model_name="event",
            name="users",
            field=models.ManyToManyField(to="tests.User"),
        ),
        migrations.AddField(
            model_name="cat",
            name="backup_home",
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.CASCADE,
                related_name="friendly_cats",
                to="tests.Location",
            ),
        ),
        migrations.AddField(
            model_name="cat",
            name="home",
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.CASCADE, to="tests.Location"
            ),
        ),
        migrations.AddField(
            model_name="cat",
            name="hunting_grounds",
            field=models.ManyToManyField(
                related_name="annoying_cats",
                related_query_name=b"getoffmylawn",
                to="tests.Location",
            ),
        ),
        migrations.AddField(
            model_name="cat",
            name="parent",
            field=models.ForeignKey(
                blank=True,
                null=True,
                on_delete=django.db.models.deletion.CASCADE,
                related_name="kittens",
                to="tests.Cat",
            ),
        ),
        migrations.AddField(
            model_name="c",
            name="d",
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.CASCADE, to="tests.D"
            ),
        ),
    ]



"""Benchmark models."""
from django.db import models


class User(models.Model):
    """User model."""

    name = models.TextField()
    groups = models.ManyToManyField("Group", related_name="users")
    created = models.DateTimeField(auto_now_add=True)
    updated = models.DateTimeField(auto_now=True)


class Group(models.Model):
    """Group model."""

    name = models.TextField()
    max_size = models.PositiveIntegerField()
    permissions = models.ManyToManyField("Permission", related_name="groups")
    created = models.DateTimeField(auto_now_add=True)
    updated = models.DateTimeField(auto_now=True)


class Permission(models.Model):
    """Permission model."""

    name = models.TextField()
    created = models.DateTimeField(auto_now_add=True)
    updated = models.DateTimeField(auto_now=True)



"""Sorting filter."""
from __future__ import annotations

from typing import TYPE_CHECKING

from django.core.exceptions import ImproperlyConfigured
from django.db.models import QuerySet
from rest_framework.exceptions import ValidationError
from rest_framework.filters import OrderingFilter
from rest_framework.request import Request
from rest_framework.serializers import SerializerMetaclass

from dynamic_rest.fields import DynamicRelationField

if TYPE_CHECKING:
    from dynamic_rest.viewsets import DynamicModelViewSet


class DynamicSortingFilter(OrderingFilter):
    """Subclass of DRF's OrderingFilter.

    This class adds support for multi-field ordering and rewritten fields.
    """

    def filter_queryset(
        self, request: Request, queryset: QuerySet, view: "DynamicModelViewSet"
    ) -> QuerySet:
        """Filter the queryset, applying the ordering.

        The `ordering_param` can be overwritten here.
        In DRF, the ordering_param is 'ordering', but we support changing it
        to allow the viewset to control the parameter.
        """
        self.ordering_param = view.SORT
        if ordering := self.get_ordering(request, queryset, view):
            queryset = queryset.order_by(*ordering)
            if any("__" in o for o in ordering):
                # add distinct() to remove duplicates
                # in case of order-by-related
                queryset = queryset.distinct()
        return queryset

    def get_ordering(
        self, request: Request, queryset: QuerySet, view: "DynamicModelViewSet"
    ):
        """Return an ordering for a given request.

        DRF expects a comma separated list, while DREST expects an array.
        This method overwrites the DRF default, so it can parse the array.
        """
        if params := view.get_request_feature(view.SORT):
            fields = [param.strip() for param in params]
            valid_ordering, invalid_ordering = self.remove_invalid_fields(
                queryset, fields, view, request
            )

            # if any of the sort fields are invalid, throw an error.
            # else return the ordering
            if invalid_ordering:
                raise ValidationError(f"Invalid filter field: {invalid_ordering}")
            else:
                return valid_ordering

        # No sorting was included
        return self.get_default_ordering(view)

    def remove_invalid_fields(
        self,
        queryset: QuerySet,
        fields: list[str],
        view: "DynamicModelViewSet",
        _: Request,
    ) -> tuple[list[str], list[str]]:
        """Remove invalid fields from an ordering.

        Overwrites the DRF default remove_invalid_fields method to return
        both the valid orderings and any invalid orderings.
        """
        valid_orderings = []
        invalid_orderings = []

        # for each field sent down from the query param,
        # determine if its valid or invalid
        for term in fields:
            stripped_term = term.lstrip("-")
            # add back the '-' add the end if necessary
            reverse_sort_term = "" if len(stripped_term) is len(term) else "-"
            if ordering := self.ordering_for(stripped_term, view):
                valid_orderings.append(reverse_sort_term + ordering)
            else:
                invalid_orderings.append(term)

        return valid_orderings, invalid_orderings

    def ordering_for(self, term: str, view: "DynamicModelViewSet") -> str | None:
        """Override DRF's ordering_for method to support rewritten fields.

        Return ordering (model field chain) for term (serializer field chain)
        or None if invalid

        Raise ImproperlyConfigured if serializer_class not set on view
        """
        if not self._is_allowed_term(term, view):
            return None

        serializer = self._get_serializer_class(view)()
        serializer_chain = term.split(".")

        model_chain = []

        for segment in serializer_chain[:-1]:
            field = serializer.get_all_fields().get(segment)

            if not (
                field
                and field.source != "*"
                and isinstance(field, DynamicRelationField)
            ):
                return None

            model_chain.append(field.source or segment)

            serializer = field.serializer_class()

        last_segment = serializer_chain[-1]
        last_field = serializer.get_all_fields().get(last_segment)

        if not last_field or last_field.source == "*":
            return None

        model_chain.append(last_field.source or last_segment)

        return "__".join(model_chain)

    def _is_allowed_term(self, term: str, view: "DynamicModelViewSet") -> bool:
        """Check if a term is allowed to be ordered on."""
        valid_fields = getattr(view, "ordering_fields", self.ordering_fields)
        all_fields_allowed = valid_fields is None or valid_fields == "__all__"

        return all_fields_allowed or term in valid_fields

    def _get_serializer_class(self, view: "DynamicModelViewSet") -> SerializerMetaclass:
        """Get the serializer class from the view."""
        # prefer the overriding method
        if hasattr(view, "get_serializer_class"):
            try:
                serializer_class = view.get_serializer_class()
            except AssertionError:
                # Raised by the default implementation if
                # no serializer_class was found
                serializer_class = None
        # use the attribute
        else:
            serializer_class = getattr(view, "serializer_class", None)

        # neither a method nor an attribute has been specified
        if serializer_class is None:
            raise ImproperlyConfigured(
                f"Cannot use {self.__class__.__name__} on a view which does"
                " not have either a 'serializer_class' or an overriding "
                "'get_serializer_class'."
            )
        return serializer_class



"""This module contains response processors."""
from collections import defaultdict

from rest_framework.serializers import ListSerializer
from rest_framework.utils.serializer_helpers import ReturnDict

from dynamic_rest.conf import settings
from dynamic_rest.tagged import TaggedDict

POST_PROCESSORS = {}


def register_post_processor(func):
    """Register a post processor function.

    Register a post processor function to be run as the final step in
    serialization. The data passed in will already have gone through the
    sideloading processor.

    Usage:
        @register_post_processor
        def my_post_processor(data):
            # do stuff with `data`
            return data
    """
    key = func.__name__
    POST_PROCESSORS[key] = func
    return func


def post_process(data):
    """Apply registered post-processors to data."""
    for post_processor in POST_PROCESSORS.values():
        data = post_processor(data)

    return data


class SideloadingProcessor(object):
    """A processor that side-loads serializer data.

    Side-loaded records are returned under top-level
    response keys and produces responses that are
    typically smaller than their nested equivalent.
    """

    def __init__(self, serializer, data):
        """Initializes and runs the processor.

        Arguments:
            serializer: a DREST serializer
            data: the serializer's representation
        """
        if isinstance(serializer, ListSerializer):
            serializer = serializer.child
        self.data = defaultdict(list)
        self.seen = defaultdict(set)
        self.plural_name = serializer.get_plural_name()
        self.name = serializer.get_name()

        # process the data, optionally sideloading
        self.process(data)

        # add the primary resource data into the response data
        resource_name = self.name if isinstance(data, dict) else self.plural_name
        self.data[resource_name] = data

    @staticmethod
    def is_dynamic(data):
        """Check whether the given data dictionary is a DREST structure.

        Arguments:
            data: A dictionary representation of a DRF serializer.
        """
        return isinstance(data, TaggedDict)

    def process(self, obj, parent=None, parent_key=None, depth=0):
        """Recursively process the data for sideloading.

        Converts the nested representation into a side-loaded representation.
        """
        if isinstance(obj, list):
            for key, o in enumerate(obj):
                # traverse into lists of objects
                self.process(o, parent=obj, parent_key=key, depth=depth)
        elif isinstance(obj, dict):
            dynamic = self.is_dynamic(obj)
            returned = isinstance(obj, ReturnDict)
            if not dynamic and not returned:
                return
            # recursively check all fields
            for key, o in obj.items():
                if isinstance(o, (list, dict)):
                    # lists or dicts indicate a relation
                    self.process(o, parent=obj, parent_key=key, depth=depth + 1)

            if not dynamic or getattr(obj, "embed", False):
                return

            serializer = obj.serializer
            name = serializer.get_plural_name()
            instance = getattr(obj, "instance", serializer.instance)
            instance_pk = instance.pk if instance else None
            pk = getattr(obj, "pk_value", instance_pk) or instance_pk

            # For polymorphic relations, `pk` can be a dict, so use the
            # string representation (dict isn't hashable).
            pk_key = repr(pk)

            # sideloading
            seen = True
            seen_set = self.seen[name]
            # if this object has not yet been seen
            if pk_key not in seen_set:
                seen = False
                seen_set.add(pk_key)

            # prevent sideloading the primary objects
            if depth == 0:
                return

            # TODO: spec out the exact behavior for secondary instances of
            # the primary resource

            # if the primary resource is embedded, add it to a prefixed key
            if name == self.plural_name:
                name = f"{settings.ADDITIONAL_PRIMARY_RESOURCE_PREFIX}{name}"

            if not seen:
                # allocate a top-level key in the data for this resource
                # type

                # move the object into a new top-level bucket
                # and mark it as seen
                self.data[name].append(obj)
            else:
                # obj side-loaded, but maybe with other fields
                for o in self.data[name]:
                    if o.instance.pk == pk:
                        o.update(obj)
                        break

            # replace the object with a reference
            if parent is not None and parent_key is not None:
                parent[parent_key] = pk



"""Utilities for dynamic_rest."""
from functools import lru_cache

from django.contrib.contenttypes.models import ContentType
from django.db import models
from django.utils.module_loading import import_string
from sqids import Sqids

FALSEY_STRINGS = (
    "0",
    "false",
    "",
)


@lru_cache()
def is_truthy(x):
    """Return True if x is truthy, False otherwise."""
    if isinstance(x, str):
        return x.lower() not in FALSEY_STRINGS
    return bool(x)


def unpack(content):
    """Unpack a content dict from a DynamicSerializer."""
    if not content:
        # empty values pass through
        return content

    keys = [k for k in content.keys() if k != "meta"]
    unpacked = content[keys[0]]
    return unpacked


def external_id_from_model_and_internal_id(model, internal_id):
    """Return a hash for the model and internal ID combination."""
    sqids = Sqids()

    if sqids is None:
        raise AssertionError(
            "To use hashids features you must set "
            "ENABLE_HASHID_FIELDS to true in your dynamic_rest settings."
        )
    return sqids.encode([ContentType.objects.get_for_model(model).id, internal_id])


def internal_id_from_model_and_external_id(model, external_id):
    """Return the internal ID from the external ID and model combination.

    Because the SqID is a combination of the model's content type and the
    internal ID, we validate here that the external ID decodes as expected,
    and that the content type corresponds to the model we're expecting.
    """
    sqids = Sqids()

    if sqids is None:
        raise AssertionError(
            "To use hashids features you must set "
            "ENABLE_HASHID_FIELDS to true in your dynamic_rest settings."
        )

    try:
        (  # pylint: disable=unbalanced-tuple-unpacking
            content_type_id,
            instance_id,
        ) = sqids.decode(external_id)
    except (TypeError, ValueError) as exc:
        raise model.DoesNotExist from exc

    content_type = ContentType.objects.get_for_id(content_type_id)

    if content_type.model_class() != model:
        raise model.DoesNotExist

    return instance_id


def model_from_definition(model_definition):
    """Return a Django model corresponding to model_definition.

    Model definition can either be a string defining how to import the model,
    or a model class.

    Arguments:
        model_definition: (str|django.db.models.Model)

    Returns:
        (django.db.models.Model)

    Implementation from
    https://github.com/evenicoulddoit/django-rest-framework-serializer-extensions
    """
    if isinstance(model_definition, str):
        model = import_string(model_definition)
    else:
        model = model_definition

    try:
        assert issubclass(model, models.Model)
    except (AssertionError, TypeError) as exc:
        raise AssertionError(f'"{model_definition}"" is not a Django model') from exc

    return model



"""Test cases for the API."""
from django.test import (  # noqa pylint: disable=unused-import
    TestCase,
    TransactionTestCase,
)
from rest_framework.test import (  # noqa pylint: disable=unused-import
    APITestCase,
    APITransactionTestCase,
)


class ResetTestCase(TransactionTestCase):
    """Reset API test case."""

    reset_sequences = True


class ResetAPITestCase(APITransactionTestCase):
    """Reset API test case."""

    reset_sequences = True



# -*- coding: utf-8 -*-
# Generated by Django 1.9 on 2016-03-10 10:52
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("tests", "0001_initial"),
    ]

    operations = [
        migrations.AddField(
            model_name="user",
            name="date_of_birth",
            field=models.DateField(blank=True, null=True),
        ),
        migrations.AlterField(
            model_name="group",
            name="name",
            field=models.TextField(unique=True),
        ),
    ]



"""Constants."""

VALID_FILTER_OPERATORS = (
    "in",
    "any",
    "all",
    "icontains",
    "contains",
    "startswith",
    "istartswith",
    "endswith",
    "iendswith",
    "year",
    "month",
    "day",
    "week_day",
    "regex",
    "range",
    "gt",
    "lt",
    "gte",
    "lte",
    "isnull",
    "eq",
    "iexact",
    None,
)



"""Generic relation field for dynamic_rest."""
from __future__ import annotations

from django.db import models
from rest_framework.exceptions import ValidationError

from dynamic_rest.fields.common import WithRelationalFieldMixin
from dynamic_rest.fields.fields import DynamicField
from dynamic_rest.routers import DynamicRouter
from dynamic_rest.tagged import TaggedDict


class DynamicGenericRelationField(WithRelationalFieldMixin, DynamicField):
    """Generic relation field for dynamic_rest."""

    def __init__(self, *args, embed=False, **kwargs):
        """Initialise the DynamicGenericRelationField."""
        if "requires" in kwargs:
            raise RuntimeError(
                "DynamicGenericRelationField does not support manual"
                " overriding of 'requires'."
            )
        super().__init__(*args, **kwargs)
        self.embed = embed

    def bind(self, field_name, parent):
        """Bind the field to the parent serializer."""
        super().bind(field_name, parent)

        source = self.source or field_name

        # Inject `requires` so required fields get prefetched properly.
        # TODO: It seems like we should be able to require the type and
        #       id fields, but that seems to conflict with some internal
        #       Django magic. Disabling `.only()` by requiring '*' seem
        #       to work more reliably...
        self.requires = [f"{source}.*", "*"]

        # Get request fields to support sideloading, but disallow field
        # inclusion/exclusion.
        request_fields = self._get_request_fields_from_parent()
        if isinstance(request_fields, dict) and len(request_fields):
            raise ValidationError(
                f"{self.parent.get_name()}.{self.field_name}"
                " does not support field inclusion/exclusion"
            )
        self.request_fields = request_fields

    def id_only(self):
        """Return whether the field is id_only."""
        # For DynamicRelationFields, id_only() is a serializer responsibility
        # but for generic relations, we want IDs to be represented differently
        # and that is a field-level concern, not an object-level concern,
        # so we handle it here.
        return not self.parent.is_field_sideloaded(self.field_name)

    @staticmethod
    def get_pk_object(type_key, id_value):
        """Get the pk object."""
        return {"type": type_key, "id": id_value}

    @staticmethod
    def get_serializer_class_for_instance(instance):
        """Get the serializer class for the instance."""
        return DynamicRouter.get_canonical_serializer(
            resource_key=None, instance=instance
        )

    def to_representation(self, value):
        """Convert the instance to a representation."""
        # Find serializer for the instance
        serializer_class = self.get_serializer_class_for_instance(value)
        if not serializer_class:
            # Can't find canonical serializer! For now, just return
            # object name and ID, and hope the client knows what to do
            # with it.
            return self.get_pk_object(
                value._meta.object_name, value.pk  # pylint: disable=protected-access
            )

        # We want the pk to be represented as an object with type,
        # rather than just the ID.
        pk_value = self.get_pk_object(serializer_class.get_name(), value.pk)
        if self.id_only():
            return pk_value

        # Serialize the object. Note that request_fields is set, but
        # field inclusion/exclusion is disallowed via check in bind()
        representation = serializer_class(
            dynamic=True,
            request_fields=self.request_fields,
            context=self.context,
            embed=self.embed,
        ).to_representation(value)

        # Pass pk object that contains type and ID to TaggedDict object
        # so that Processor can use it when the field gets side-loaded.
        if isinstance(representation, TaggedDict):
            representation.pk_value = pk_value
        return representation

    def to_internal_value(self, data: dict) -> models.Model | None:
        """Convert the data to an internal value."""
        model_name = data.get("type", None)
        model_id = data.get("id", None)
        if not (model_name and model_id):
            return None

        serializer_class = DynamicRouter.get_canonical_serializer(
            resource_key=None, resource_name=model_name
        )
        if not serializer_class:
            return None
        model = serializer_class.get_model()
        return model.objects.get(id=model_id) if model else None



"""Django app config for dynamic_rest."""
from django.apps import AppConfig
from django.core.exceptions import ImproperlyConfigured

from dynamic_rest.conf import settings


class DynamicRestConfig(AppConfig):
    """Django app config for dynamic_rest."""

    name = "dynamic_rest"
    verbose_name = "Django Dynamic Rest"

    def ready(self):
        """Perform app config checks."""
        if hasattr(settings, "ENABLE_HASHID_FIELDS") and settings.ENABLE_HASHID_FIELDS:
            if not hasattr(settings, "HASHIDS_SALT") or settings.HASHIDS_SALT is None:
                raise ImproperlyConfigured(
                    "ENABLED_HASHID_FIELDS is True in your settings,"
                    "but no HASHIDS_SALT string was set!"
                )



"""This module contains custom serializer classes."""
import copy
import inspect
import os

import inflection
from django.db import models
from django.utils.functional import cached_property
from rest_framework import __version__ as drf_version
from rest_framework import exceptions
from rest_framework import fields as drf_fields
from rest_framework import serializers
from rest_framework.fields import SkipField
from rest_framework.relations import RelatedField
from rest_framework.utils.serializer_helpers import ReturnDict, ReturnList

from dynamic_rest import prefetch
from dynamic_rest.bases import (
    CacheableFieldMixin,
    DynamicSerializerBase,
    resettable_cached_property,
)
from dynamic_rest.conf import settings
from dynamic_rest.fields import DynamicGenericRelationField, DynamicRelationField
from dynamic_rest.links import merge_link_object
from dynamic_rest.meta import get_model_table
from dynamic_rest.processors import SideloadingProcessor, post_process
from dynamic_rest.tagged import TaggedDict
from dynamic_rest.utils import external_id_from_model_and_internal_id

OPTS = {"ENABLE_FIELDS_CACHE": os.environ.get("ENABLE_FIELDS_CACHE", False)}
FIELDS_CACHE = {}
DRF_VERSION = drf_version.split(".")
OLD_DRF = int(DRF_VERSION[0]) <= 3 and int(DRF_VERSION[1]) < 5


class WithResourceKeyMixin(object):
    """Mixin for serializers that have a resource key."""

    def get_resource_key(self):
        """Return canonical resource key, usually the DB table name."""
        model = self.get_model()
        if model:
            return get_model_table(model)
        return self.get_name()


class DynamicListSerializer(
    CacheableFieldMixin, WithResourceKeyMixin, serializers.ListSerializer
):
    """Custom ListSerializer class.

    This implementation delegates DREST-specific methods to
    the child serializer and performs post-processing before
    returning the data.
    """

    update_lookup_field = "id"

    def __init__(self, *args, **kwargs):
        """Initializes the serializer."""
        super().__init__(*args, **kwargs)
        self.child.parent = self

    def to_representation(self, data):
        """Delegates to the child serializer."""
        iterable = data.all() if isinstance(data, models.Manager) else data
        child = self.child
        return [child.to_representation(item) for item in iterable]

    def get_model(self):
        """Get the child's model."""
        return self.child.get_model()

    def get_name(self):
        """Get the child's name."""
        return self.child.get_name()

    def get_plural_name(self):
        """Get the child's plural name."""
        return self.child.get_plural_name()

    def id_only(self):
        """Get the child's rendering mode."""
        return self.child.id_only()

    @resettable_cached_property
    def data(self):  # pylint: disable=invalid-overridden-method
        """Get the data, after performing post-processing if necessary."""
        data = super().data
        processed_data = (
            ReturnDict(SideloadingProcessor(self, data).data, serializer=self)
            if self.child.envelope
            else ReturnList(data, serializer=self)
        )
        return post_process(processed_data)

    def update(self, queryset, validated_data):  # pylint: disable=arguments-renamed
        """Update a queryset with validated data."""
        lookup_attr = getattr(self.child.Meta, "update_lookup_field", "id")

        lookup_objects = {
            str(entry.pop(lookup_attr)): entry for entry in validated_data
        }

        lookup_keys = lookup_objects.keys()

        if not all((bool(_) and not inspect.isclass(_) for _ in lookup_keys)):
            raise exceptions.ValidationError("Invalid lookup key value.")

        # Since this method is given a queryset which can have many
        # model instances, first find all objects to update
        # and only then update the models.
        try:
            objects_to_update = queryset.filter(**{f"{lookup_attr}__in": lookup_keys})
        except Exception as exc:
            raise exceptions.ValidationError(
                f'Invalid lookup keys: {", ".join(lookup_keys)}'
            ) from exc
        keys_len = len(lookup_keys)
        object_update_count = objects_to_update.count()
        if keys_len != object_update_count:
            raise exceptions.ValidationError(
                "Could not find all objects to update: "
                f"{keys_len} != {object_update_count}."
            )

        updated_objects = []
        for object_to_update in objects_to_update:
            lookup_key = getattr(object_to_update, lookup_attr)
            lookup_key = str(lookup_key)
            data = lookup_objects.get(lookup_key)
            # Use model serializer to actually update the model
            # in case that method is overwritten.
            updated_objects.append(self.child.update(object_to_update, data))

        return updated_objects


class WithDynamicSerializerMixin(
    CacheableFieldMixin, WithResourceKeyMixin, DynamicSerializerBase
):
    """Base class for DREST serializers.

    This class provides support for dynamic field inclusions/exclusions.

    Like DRF, DREST serializers support a few Meta class options:
        - model - class
        - name - string
        - plural_name - string
        - defer_many_relations - bool
        - hash_ids - bool
        - fields - list of strings
        - deferred_fields - list of strings
        - immutable_fields - list of strings
        - read_only_fields - list of strings
        - untrimmed_fields - list of strings
    """

    ENABLE_FIELDS_CACHE = False

    def __new__(cls, *args, **kwargs):
        """
        Custom constructor.

        Sets the ListSerializer to DynamicListSerializer
        to avoid re-evaluating querysets.

        Addresses DRF 3.1.0 bug:
        https://github.com/tomchristie/django-rest-framework/issues/2704
        """
        meta = getattr(cls, "Meta", None)
        if not meta:
            meta = type("Meta", (), {})
            cls.Meta = meta

        list_serializer_class = getattr(
            meta,
            "list_serializer_class",
            settings.LIST_SERIALIZER_CLASS or DynamicListSerializer,
        )
        if not issubclass(list_serializer_class, DynamicListSerializer):
            list_serializer_class = DynamicListSerializer
        meta.list_serializer_class = list_serializer_class
        return super().__new__(cls, *args, **kwargs)

    def __init__(
        self,
        instance=None,
        data=drf_fields.empty,
        only_fields=None,
        include_fields=None,
        exclude_fields=None,
        request_fields=None,
        sideloading=None,
        debug=False,
        dynamic=True,
        embed=False,
        envelope=False,
        **kwargs,
    ):
        """
        Custom initializer that builds `request_fields`.

        Arguments:
            instance: Initial instance, used by updates.
            data: Initial data, used by updates / creates.
            only_fields: List of field names to render.
            include_fields: List of field names to include.
            exclude_fields: List of field names to exclude.
            request_fields: Map of field names that supports
                nested inclusions / exclusions.
            sideloading: If True, force sideloading for all descendents.
                If False, force embedding for all descendents.
                If None (default), respect descendents' embed parameters.
            debug: If True, include debug information in the response.
            dynamic: If False, disable inclusion / exclusion features.
            embed: If True, embed the current representation.
                If False, sideload the current representation.
            envelope: If True, wrap `.data` in an envelope.
                If False, do not use an envelope.
        """
        name = self.get_name()
        if data is not drf_fields.empty and name in data and len(data) == 1:
            # support POST/PUT key'd by resource name
            data = data[name]

        if data is not drf_fields.empty:
            # if a field is nullable but not required and the implementation
            # passes null as a value, remove the field from the data
            # this addresses the frontends that send
            # undefined resource fields as null on POST/PUT
            for field_name, field in self.get_all_fields().items():
                if (
                    field.allow_null is False
                    and field.required is False
                    and field_name in data
                    and data[field_name] is None
                ):
                    data.pop(field_name)

        kwargs["instance"] = instance
        kwargs["data"] = data

        # "sideload" argument is pending deprecation as of 1.6
        if kwargs.pop("sideload", False):
            # if "sideload=True" is passed, turn on the envelope
            envelope = True

        super().__init__(**kwargs)

        self.envelope = envelope
        self.sideloading = sideloading
        self.debug = debug
        self.dynamic = dynamic
        self.request_fields = request_fields or {}

        # `embed` is overriden by `sideloading`
        self.embed = embed if sideloading is None else not sideloading

        self._dynamic_init(only_fields, include_fields, exclude_fields)
        self.enable_optimization = settings.ENABLE_SERIALIZER_OPTIMIZATIONS
        # self.id_only = self.dynamic and self.request_fields is True

    def _dynamic_init(self, only_fields, include_fields, exclude_fields):
        """
        Modifies `request_fields` via higher-level dynamic field interfaces.

        Arguments:
            only_fields: List of field names to render.
                All other fields will be deferred (respects side-loads).
            include_fields: List of field names to include.
                Adds to default field set, (respects side-loads).
                `*` means include all fields.
            exclude_fields: List of field names to exclude.
                Removes from default field set. If set to '*', all fields are
                removed, except for ones that are explicitly included.
        """
        if not self.dynamic:
            return

        if (
            isinstance(self.request_fields, dict)
            and self.request_fields.pop("*", None) is False
        ):
            exclude_fields = "*"

        only_fields = set(only_fields or [])
        include_fields = include_fields or []
        exclude_fields = exclude_fields or []

        if only_fields:
            exclude_fields = "*"
            include_fields = only_fields

        if exclude_fields == "*":
            # First exclude all, then add back in explicitly included fields.
            include_fields = set(
                list(include_fields)
                + [
                    field
                    for field, val in self.request_fields.items()
                    if val or val == {}
                ]
            )
            all_fields = set(self.get_all_fields().keys())  # this is slow
            exclude_fields = all_fields - include_fields
        elif include_fields == "*":
            all_fields = set(self.get_all_fields().keys())  # this is slow
            include_fields = all_fields

        for name in exclude_fields:
            self.request_fields[name] = False

        for name in include_fields:
            if not isinstance(self.request_fields.get(name), dict):
                # not sideloading this field
                self.request_fields[name] = True

    @classmethod
    def get_model(cls):
        """Get the model, if the serializer has one.

        Model serializers should implement this method.
        """
        return None

    @classmethod
    def get_name(cls):
        """Get the serializer name.

        The name can be defined on the Meta class or will be generated
        automatically from the model name.
        """
        meta = cls.Meta
        if not hasattr(meta, "name"):
            class_name = getattr(cls.get_model(), "__name__", None)
            name = inflection.underscore(class_name) if class_name else None
            setattr(meta, "name", name)
            return name

        return meta.name

    @classmethod
    def get_plural_name(cls):
        """Get the serializer's plural name.

        The plural name may be defined on the Meta class.
        If the plural name is not defined,
        the pluralized form of the name will be returned.
        """
        if not hasattr(cls.Meta, "plural_name"):
            setattr(cls.Meta, "plural_name", inflection.pluralize(cls.get_name()))
        return cls.Meta.plural_name

    def get_request_attribute(self, attribute, default=None):
        """Get an attribute from the request object."""
        return getattr(self.context.get("request"), attribute, default)

    def get_request_method(self):
        """Get the request method."""
        return self.get_request_attribute("method", "").upper()

    @resettable_cached_property
    def _all_fields(self):
        """Returns the entire serializer field set.

        Does not respect dynamic field inclusions/exclusions.
        """
        clazz = self.__class__
        if (
            not settings.ENABLE_FIELDS_CACHE
            or not self.ENABLE_FIELDS_CACHE
            or clazz not in FIELDS_CACHE
        ):
            all_fields = super().get_fields()

            if settings.ENABLE_FIELDS_CACHE and self.ENABLE_FIELDS_CACHE:
                FIELDS_CACHE[clazz] = all_fields
        else:
            all_fields = copy.copy(FIELDS_CACHE[clazz])
            for k, field in all_fields.items():
                if hasattr(field, "reset"):
                    field.reset()

        for k, field in all_fields.items():
            field.field_name = k
            field.parent = self

        return all_fields

    def get_all_fields(self):
        """Returns the entire serializer field set."""
        return self._all_fields

    def _get_flagged_field_names(self, fields, attr, meta_attr=None):
        """Returns a set of field names that have a given attribute set."""
        if meta_attr is None:
            meta_attr = f"{attr}_fields"
        meta_list = set(getattr(self.Meta, meta_attr, []))
        return {
            name
            for name, field in fields.items()
            if getattr(field, attr, None) is True or name in meta_list
        }

    def _get_deferred_field_names(self, fields):
        """Returns a set of field names that are deferred."""
        deferred_fields = self._get_flagged_field_names(fields, "deferred")
        defer_many_relations = (
            settings.DEFER_MANY_RELATIONS
            if not hasattr(self.Meta, "defer_many_relations")
            else self.Meta.defer_many_relations
        )
        if defer_many_relations:
            # Auto-defer all fields, unless the 'deferred' attribute
            # on the field is specifically set to False.
            many_fields = self._get_flagged_field_names(fields, "many")
            deferred_fields.update(
                {
                    name
                    for name in many_fields
                    if getattr(fields[name], "deferred", None) is not False
                }
            )

        return deferred_fields

    def flag_fields(self, all_fields, fields_to_flag, attr, value):
        """Flag fields."""
        for name in fields_to_flag:
            field = all_fields.get(name)
            if not field:
                continue
            setattr(field, attr, value)

    def get_fields(self):
        """Returns the serializer's field set.

        If `dynamic` is True, respects field inclusions/exclusions.
        Otherwise, reverts back to standard DRF behavior.
        """
        all_fields = self.get_all_fields()
        if self.dynamic is False:
            return all_fields

        if self.id_only():
            return {}

        serializer_fields = copy.deepcopy(all_fields)
        request_fields = self.request_fields
        deferred = self._get_deferred_field_names(serializer_fields)

        # apply request overrides
        if request_fields:
            for name, include in request_fields.items():
                if name not in serializer_fields:
                    raise exceptions.ParseError(
                        f'"{name}" is not a valid field name for "{self.get_name()}".'
                    )
                if include is not False and name in deferred:
                    deferred.remove(name)
                elif include is False:
                    deferred.add(name)

        for name in deferred:
            serializer_fields.pop(name)

        # Set read_only flags based on read_only_fields meta list.
        # Here to cover DynamicFields not covered by DRF.
        ro_fields = getattr(self.Meta, "read_only_fields", [])
        self.flag_fields(serializer_fields, ro_fields, "read_only", True)

        pw_fields = getattr(self.Meta, "untrimmed_fields", [])
        self.flag_fields(
            serializer_fields,
            pw_fields,
            "trim_whitespace",
            False,
        )

        # Toggle read_only flags for immutable fields.
        # Note: This overrides `read_only` if both are set, to allow
        #       inferred DRF fields to be made immutable.
        immutable_field_names = self._get_flagged_field_names(
            serializer_fields, "immutable"
        )
        self.flag_fields(
            serializer_fields,
            immutable_field_names,
            "read_only",
            value=self.get_request_method() != "POST",
        )

        return serializer_fields

    def is_field_sideloaded(self, field_name):
        """Check if a field is side-loaded."""
        if not isinstance(self.request_fields, dict):
            return False
        return isinstance(self.request_fields.get(field_name), dict)

    def get_link_fields(self):
        """Return a dict of linkable fields."""
        return self._link_fields

    @resettable_cached_property
    def _link_fields(self):
        """Construct dict of name:field for linkable fields."""
        query_params = self.get_request_attribute("query_params", {})
        if "exclude_links" in query_params:
            return {}
        else:
            all_fields = self.get_all_fields()
            return {
                name: field
                for name, field in all_fields.items()
                if isinstance(field, DynamicRelationField)
                and getattr(field, "link", True)
                and not (
                    # Skip sideloaded fields
                    name in self.fields
                    and self.is_field_sideloaded(name)
                )
                and not (
                    # Skip included single relations
                    # TODO: Use links, when we can generate canonical URLs
                    name in self.fields
                    and not getattr(field, "many", False)
                )
            }

    @cached_property
    def _readable_fields(self):
        """Return a list of readable fields."""
        # NOTE: Copied from DRF, exists in 3.2.x but not 3.1
        return [field for field in self.fields.values() if not field.write_only]

    @cached_property
    def _readable_id_fields(self):
        """Return a list of readable id fields."""
        fields = self._readable_fields
        return {
            field
            for field in fields
            if (
                isinstance(field, (DynamicRelationField, RelatedField))
                and not isinstance(self.request_fields.get(field.field_name), dict)
            )
        }

    def _get_hash_ids(self):
        """
        Check whether ids should be hashed or not.

        Determined by the hash_ids boolean Meta field.
        Defaults to False.

        Returns:
            Boolean.
        """
        if hasattr(self.Meta, "hash_ids"):
            return self.Meta.hash_ids
        else:
            return False

    def _faster_to_representation(self, instance):
        """Modified to_representation with optimizations.

        1) Returns a plain old dict as opposed to OrderedDict.
            (Constructing ordered dict is ~100x slower than `{}`.)
        2) Ensure we use a cached list of fields
            (this optimization exists in DRF 3.2 but not 3.1)

        Arguments:
            instance: a model instance or data object
        Returns:
            Dict of primitive datatypes.
        """
        ret = {}
        fields = self._readable_fields

        is_fast = isinstance(instance, prefetch.FastObject)
        id_fields = self._readable_id_fields
        class_name = self.__class__.__name__
        for field in fields:
            field_name = field.field_name
            field_source = field.source
            # we exclude dynamic fields here because the proper FastQuery
            # de-referencing happens in the `get_attribute` method now
            if is_fast and not isinstance(
                field, (DynamicGenericRelationField, DynamicRelationField)
            ):
                if field in id_fields and field_source not in instance:
                    # TODO - make better.
                    attribute = instance.get(f"{field_source}_id")
                    ret[field_name] = attribute
                    continue
                else:
                    try:
                        attribute = instance[field_source]
                    except KeyError:
                        # slower, but does more stuff
                        # Also, some temp debugging
                        if hasattr(instance, field_source):
                            attribute = getattr(instance, field_source)
                        else:
                            # Fall back on DRF behavior
                            attribute = field.get_attribute(instance)
                            print(f"Missing {field_name} from {class_name}")
            else:
                try:
                    attribute = field.get_attribute(instance)
                except SkipField:
                    continue

            if attribute is None:
                # We skip `to_representation` for `None` values so that
                # fields do not have to explicitly deal with that case.
                ret[field_name] = None
            else:
                ret[field_name] = field.to_representation(attribute)

        return ret

    @resettable_cached_property
    def obj_cache(self):
        """Cache for objects."""
        # Note: This gets cached by resettable_cached_property so this
        #       function only needs to return the initial value.
        return {}

    def _to_representation(self, instance):
        """Uncached `to_representation`."""
        if self.enable_optimization:
            representation = self._faster_to_representation(instance)
        else:
            representation = super().to_representation(instance)

        if settings.ENABLE_LINKS:
            # TODO: Make this function configurable to support other
            #       formats like JSON API link objects.
            representation = merge_link_object(self, representation, instance)

        if self.debug:
            representation["_meta"] = {
                "id": instance.pk,
                "type": self.get_plural_name(),
            }

        # tag the representation with the serializer and instance
        return TaggedDict(
            representation, serializer=self, instance=instance, embed=self.embed
        )

    def to_representation(self, instance):
        """Modified to_representation method. Optionally may cache objects.

        Arguments:
            instance: A model instance or data object.
        Returns:
            Instance ID if the serializer is meant to represent its ID.
            Otherwise, a tagged data dict representation.
        """
        if self.id_only():
            if self._get_hash_ids():
                return external_id_from_model_and_internal_id(
                    self.get_model(), instance.pk
                )
            return instance.pk

        pk = getattr(instance, "pk", None)
        representation = self._to_representation(instance)
        if not settings.ENABLE_SERIALIZER_OBJECT_CACHE or pk is None:
            return representation
        if pk not in self.obj_cache:
            self.obj_cache[pk] = representation
        return self.obj_cache[pk]

    def to_internal_value(self, data):
        """Modified to_internal_value method."""
        value = super().to_internal_value(data)
        id_attr = getattr(self.Meta, "update_lookup_field", "id")
        request_method = self.get_request_method()

        # Add update_lookup_field field back to validated data
        # since super by default strips out read-only fields
        # hence id will no longer be present in validated_data.
        if all(
            (
                isinstance(self.root, DynamicListSerializer),
                id_attr,
                request_method in ("PUT", "PATCH"),
            )
        ):
            id_field = self.fields[id_attr]
            id_value = id_field.get_value(data)
            value[id_attr] = id_value

        return value

    def save(self, *args, **kwargs):
        """Serializer save that address prefetch issues."""
        update = getattr(self, "instance", None) is not None
        instance = super().save(*args, **kwargs)
        view = self._context.get("view")
        if view and update:
            if OLD_DRF:
                # Reload the object on update
                # to get around prefetch cache issues
                # Fixed in DRF in 3.5.0
                instance = self.instance = view.get_object()
        return instance

    def id_only(self):
        """Whether the serializer should return an ID instead of an object.

        Returns:
            True if and only if `request_fields` is True.
        """
        return self.dynamic and self.request_fields is True

    @resettable_cached_property
    def data(self):
        """Get the data, after performing post-processing if necessary."""
        if hasattr(self, "_processed_data"):
            return self._processed_data
        data = super().data
        data = SideloadingProcessor(self, data).data if self.envelope else data
        processed_data = ReturnDict(data, serializer=self)
        self._processed_data = data = post_process(processed_data)
        return data


class WithDynamicModelSerializerMixin(WithDynamicSerializerMixin):
    """Adds DREST serializer methods specific to model-based serializers."""

    @classmethod
    def get_model(cls):
        """Get the model, if the serializer has one."""
        return getattr(cls.Meta, "model", None)

    def get_id_fields(self):
        """Get the list of ID fields.

        Called to return a list of fields consisting of, at minimum,
        the PK field name. The output of this method is used to
        construct a Prefetch object with a .only() queryset
        when this field is not being side-loaded, but we need to
        return a list of IDs.
        """
        model = self.get_model()
        # get PK field name
        out = [model._meta.pk.name]  # pylint: disable=protected-access

        # If this is being called, it means it
        # is a many-relation  to its parent.
        # Django wants the FK to the parent,
        # but since accurately inferring the FK
        # pointing back to the parent is less than trivial,
        # we will just pull all ID fields.
        # TODO: We also might need to return all non-nullable fields,
        #    or else it is possible Django will issue another request.
        for field in model._meta.fields:  # pylint: disable=protected-access
            if isinstance(field, models.ForeignKey):
                out.append(f"{field.name}_id")

        return out


class DynamicModelSerializer(
    WithDynamicModelSerializerMixin, serializers.ModelSerializer
):
    """DREST-compatible model-based serializer."""

    pass


class EphemeralObject(object):
    """Object that initializes attributes from a dict."""

    def __init__(self, values_dict):
        """Initialize EphemeralObject."""
        if "pk" not in values_dict:
            raise RuntimeError('"pk" key is required')
        self.__dict__.update(values_dict)


class DynamicEphemeralSerializer(WithDynamicSerializerMixin, serializers.Serializer):
    """DREST-compatible baseclass for non-model serializers."""

    def to_representation(self, instance):
        """Serialize instance.

        Provides post-processing. Subclasses should implement their own
        to_representation method, but pass the resulting dict through
        this function to get tagging and field selection.

        Arguments:
            instance: Serialized dict, or object. If an object,
                it will be serialized by the super class's
                to_representation() method.
        """
        if not isinstance(instance, dict):
            data = super().to_representation(instance)
        else:
            data = instance
            instance = EphemeralObject(data)

        if self.id_only():
            return data
        return TaggedDict(data, serializer=self, instance=instance)



"""This module contains utilities to support API links."""
from dynamic_rest.conf import settings
from dynamic_rest.routers import DynamicRouter


def merge_link_object(serializer, data, instance):
    """Add a 'links' attribute to the data that maps field names to URLs.

    NOTE: This is the format that Ember Data supports, but alternative
          implementations are possible to support other formats.
    """
    link_object = {}

    if not getattr(instance, "pk", None):
        # If instance doesn't have a `pk` field, we'll assume it doesn't
        # have a canonical resource URL to hang a link off of.
        # This generally only affects Ephemeral Objects.
        return data

    link_fields = serializer.get_link_fields()
    for name, field in link_fields.items():
        # For included fields, omit link if there's no data.
        if name in data and not data[name]:
            continue

        link = getattr(field, "link", None)
        if link is None:
            base_url = ""
            if settings.ENABLE_HOST_RELATIVE_LINKS:
                # if the resource isn't registered, this will default back to
                # using resource-relative urls for links.
                base_url = (
                    DynamicRouter.get_canonical_path(
                        serializer.get_resource_key(), instance.pk
                    )
                    or ""
                )
            link = f"{base_url}{name}/"
        # Default to DREST-generated relation endpoints.
        elif callable(link):
            link = link(name, field, data, instance)

        link_object[name] = link

    if link_object:
        data["links"] = link_object
    return data



"""Fast Filter Backend."""
from django.db.models import Model, QuerySet

from dynamic_rest.filters.base import DynamicFilterBackend
from dynamic_rest.prefetch import FastPrefetch, FastQuery
from dynamic_rest.serializers import DynamicModelSerializer


class FastDynamicFilterBackend(DynamicFilterBackend):
    """A DRF filter backend that constructs DREST querysets."""

    def _create_prefetch(self, source: str, queryset: QuerySet) -> FastPrefetch:
        """Create a Prefetch object."""
        return FastPrefetch(source, queryset=queryset)

    def _get_queryset(
        self, queryset: QuerySet = None, serializer: DynamicModelSerializer = None
    ) -> FastQuery:
        """Get the base queryset for this request."""
        queryset = super()._get_queryset(queryset=queryset, serializer=serializer)

        if not isinstance(queryset, FastQuery):
            queryset = FastQuery(queryset)

        return queryset

    def _make_model_queryset(self, model: Model) -> FastQuery:
        """Make a queryset for a model."""
        queryset = super()._make_model_queryset(model)
        return FastQuery(queryset)

    def _serializer_filter(
        self, serializer: DynamicModelSerializer, queryset: QuerySet
    ) -> QuerySet:
        """Filter a queryset using a serializer's filter_queryset method."""
        queryset.queryset = serializer.filter_queryset(queryset.queryset)
        return queryset



"""Prefetching for FastQuery."""
import copy
import traceback
from collections import defaultdict
from functools import lru_cache

from django.db import models
from django.db.models import Prefetch, QuerySet

from dynamic_rest.meta import get_model_field_and_type


class FastObject(dict):
    """FastObject is a dict-like object that allows for dot notation."""

    def __init__(self, *args, **kwargs):
        """Initialise the FastObject."""
        self.pk_field = kwargs.pop("pk_field", "id")
        super().__init__(*args)

    @property
    def pk(self):
        """Return the pk."""
        return self[self.pk_field]

    @lru_cache()
    def _slow_getattr(self, name):
        """Get an attribute."""
        if "." in name:
            parts = name.split(".")
            obj = self
            for part in parts:
                obj = obj[part]
            return obj
        elif name == "*":
            return self
        else:
            raise AttributeError(name)

    def __getattr__(self, name):
        """Get an attribute."""
        try:
            return self[name]
        except KeyError:
            # Fast approach failed, fall back on slower logic.
            return self._slow_getattr(name)

    def __setattr__(self, name, value):
        """Set an attribute."""
        if name not in ("pk_field", "pk"):
            self[name] = value
        else:
            super().__setattr__(name, value)


class SlowObject(dict):
    """SlowObject is a dict-like object that allows for dot notation."""

    def __init__(self, *args, slow_object=None, **kwargs):
        """Initialise the SlowObject."""
        self.pk_field = kwargs.pop("pk_field", "id")
        self.data = slow_object
        super().__init__(slow_object=slow_object.__dict__, *args)

    @property
    def pk(self):
        """Return the pk."""
        return self[self.pk_field]

    def __getitem__(self, value):
        """Get an item."""
        if hasattr(self.data, str(value)):
            return getattr(self.data, str(value))

        # for the purpose of mapping serialized model + '_id' fields back to
        # internal models, we need to check if that pattern is present
        is_nested_obj = value.split("_")
        test_attr = "_".join(is_nested_obj[:-1])
        attr_exists = hasattr(self.data, test_attr)
        if is_nested_obj[-1] == "id" and attr_exists:
            return getattr(self.data, test_attr).id

        return None

    def __iter__(self):
        """Iterate over the object."""
        return iter([self.data])

    def __getattr__(self, value):
        """Get an attribute."""
        # EAFP
        return getattr(self.data, str(value))


class FastList(list):
    """FastList is a list-like object that allows for dot notation."""

    # shim for related m2m record sets
    def all(self):
        """Return self."""
        return self


class FastPrefetch(object):
    """FastPrefetch is a prefetch object that allows for dot notation."""

    def __init__(self, field, queryset=None):
        """Initialise the FastPrefetch."""
        if isinstance(queryset, models.Manager):
            queryset = queryset.all()
        if isinstance(queryset, QuerySet):
            queryset = FastQuery(queryset)

        assert queryset is None or isinstance(queryset, FastQuery)

        self.field = field
        self.query = queryset

    @classmethod
    def make_from_field(cls, model=None, field_name=None, field=None):
        """Create a FastPrefetch from a field."""
        assert (
            model and field_name
        ) or field, "make_from_field required model+field_name or field"

        # For nested prefetch, only handle first level.
        field_parts = field_name.split("__")
        field_name = field_parts[0]
        nested_prefetches = "__".join(field_parts[1:])

        field, ftype = get_model_field_and_type(model, field_name)
        if not ftype:
            raise RuntimeError(f"{field_name} is not prefetchable")

        qs = field.remote_field.model.objects.all()

        field_name = field_name or field.name
        prefetch = cls(field_name, qs)

        # For nested prefetch, recursively pass down remainder
        if nested_prefetches:
            prefetch.query.prefetch_related(nested_prefetches)

        return prefetch

    @classmethod
    def make_from_prefetch(cls, prefetch: Prefetch, parent_model):
        """Create a FastPrefetch from a Django Prefetch object."""
        assert isinstance(prefetch, Prefetch)

        if isinstance(prefetch.queryset, FastQuery):
            return cls(prefetch.prefetch_through, prefetch.queryset)
        else:
            return cls.make_from_field(
                model=parent_model, field_name=prefetch.prefetch_through
            )


class FastQueryCompatMixin(object):
    """Compatibility mixin for FastQuery.

    Mixins for FastQuery to provide QuerySet-compatibility APIs.
    They basically just modify the underlying QuerySet object.
    Separated in a mixin, so it's clearer which APIs are supported.
    """

    def prefetch_related(self, *args):
        """Prefetch related objects."""
        model = self.model
        try:
            for arg in args:
                if isinstance(arg, str):
                    arg = FastPrefetch.make_from_field(model=model, field_name=arg)
                elif isinstance(arg, Prefetch):
                    arg = FastPrefetch.make_from_prefetch(arg, model)
                if not isinstance(arg, FastPrefetch):
                    raise ValueError("Must be FastPrefetch object")

                if arg.field in self.prefetches:
                    raise RuntimeError(
                        f"Prefetch for field '{arg.field}' already exists."
                    )
                self.prefetches[arg.field] = arg
        except Exception:  # noqa pylint: disable=broad-exception-caught
            traceback.print_exc()

        return self

    def only(self, *fields):  # pylint: disable=unused-argument
        """Only support not implemented.

        # TODO: support this for realz
        self.fields = set(self.fields) + set(fields)
        """
        return self

    def exclude(self, *args, **kwargs):
        """Exclude query args."""
        self.queryset = self.queryset.exclude(*args, **kwargs)
        return self

    def count(self):
        """Return the count."""
        qs = self.queryset._clone()  # pylint: disable=protected-access
        return qs.count()

    def extra(self, *args, **kwargs):
        """Extra query args."""
        self.queryset = self.queryset.extra(*args, **kwargs)
        return self

    def filter(self, *args, **kwargs):
        """Filter query args."""
        self.queryset = self.queryset.filter(*args, **kwargs)
        return self

    def order_by(self, *ordering):
        """Order by query args."""
        self.queryset = self.queryset.order_by(*ordering)
        return self

    def distinct(self, *args, **kwargs):
        """Distinct query args."""
        self.queryset = self.queryset.distinct(*args, **kwargs)
        return self

    def get(self, *args, **kwargs):
        """Get query args."""
        # Returns ORM object
        queryset = self._get_django_queryset()
        return queryset.get(*args, **kwargs)

    def first(self, *args, **kwargs):  # pylint: disable=unused-argument
        """Get first query args.

        Returns ORM object.
        """
        queryset = self._get_django_queryset()
        return queryset.first()

    @property
    def query(self):
        """Return the query."""
        return self.queryset.query

    def _clone(self):
        """Clone the queryset."""
        new = copy.copy(self)
        new.queryset = new.queryset._clone()  # pylint: disable=protected-access
        return new

    def _get_django_queryset(self):
        """Return Django QuerySet with prefetches properly configured."""
        prefetches = []
        for field, fprefetch in self.prefetches.items():
            has_query = hasattr(fprefetch, "query")
            qs = fprefetch.query.queryset if has_query else None
            prefetches.append(Prefetch(field, queryset=qs))

        queryset = self.queryset
        if prefetches:
            queryset = queryset.prefetch_related(*prefetches)

        return queryset

    def annotate(self, *args, **kwargs):
        """Annotate query args."""
        self.queryset = self.queryset.annotate(*args, **kwargs)
        return self


class FastQuery(FastQueryCompatMixin, object):
    """FastQuery is a fast queryset that supports prefetching."""

    def __init__(self, queryset):
        """Initialise the FastQuery."""
        if isinstance(queryset, models.Manager):
            queryset = queryset.all()
        self.queryset = queryset
        self.model = queryset.model
        self.prefetches = {}
        self.fields = None
        self.pk_field = queryset.model._meta.pk.attname
        self._data = None
        self._my_ids = None

    def execute(self):
        """Execute."""
        if self._data is not None:
            return self._data

        # TODO: check if queryset already has values() called
        # TODO: use self.fields
        qs = self.queryset._clone()  # pylint: disable=protected-access

        use_fastquery = getattr(self.model, "USE_FASTQUERY", True)

        if use_fastquery:
            data = list(qs.values())

            self.merge_prefetch(data)
            self._data = FastList(
                map(lambda obj: FastObject(obj, pk_field=self.pk_field), data)
            )
        else:

            def make_prefetch(fast_prefetch):
                """Make a prefetch."""
                queryset = None
                if fast_prefetch.query is not None:
                    queryset = fast_prefetch.query.queryset
                return Prefetch(fast_prefetch.field, queryset=queryset)

            prefetches = [
                make_prefetch(prefetch) for prefetch in self.prefetches.values()
            ]
            if prefetches:
                qs = qs.prefetch_related(*prefetches)
            self._data = FastList(
                map(lambda obj: SlowObject(obj, pk_field=self.pk_field), qs.all())
            )

        return self._data

    def __iter__(self):
        """Allow this to be cast to an iterable.

        Note: as with Django QuerySets, calling this will cause the
              query to execute.
        """
        return iter(self.execute())

    def __getitem__(self, k):
        """Support list index and slicing, similar to Django QuerySet."""
        if self._data is not None:
            # Query has already been executed. Extract from local cache.
            return self._data[k]

        # Query hasn't yet been executed. Update queryset.
        if isinstance(k, slice):
            if k.step:
                raise TypeError("Stepping not supported")

            start = int(k.start) if k.start is not None else None
            stop = int(k.stop) if k.stop is not None else None
            self.queryset.query.set_limits(start, stop)
        else:
            self.queryset.query.set_limits(k, k + 1)
        return self.execute()

    def __len__(self):
        """Return the length."""
        return len(self.execute())

    def get_ids(self, ids):
        """Return a queryset filtered by IDs."""
        self.queryset = self.queryset.filter(pk__in=ids)
        return self

    def merge_prefetch(self, data):
        """Merge prefetches into data."""
        model = self.queryset.model

        rel_func_map = {
            "fk": self.merge_fk,
            "o2o": self.merge_o2o,
            "o2or": self.merge_o2or,
            "m2m": self.merge_m2m,
            "m2o": self.merge_m2o,
        }

        for prefetch in self.prefetches.values():
            # TODO: here we assume we're dealing with Prefetch objects
            #       we could support field notation as well.
            field, rel_type = get_model_field_and_type(model, prefetch.field)
            if not rel_type:
                # Not a relational field... weird.
                # TODO: maybe raise?
                continue

            func = rel_func_map[rel_type]
            func(data, field, prefetch)

        return data

    def _make_id_map(self, items, pk_field="id"):
        """Make an ID map."""
        return {item[pk_field]: item for item in items}

    def _get_my_ids(self, data):
        """Get my IDs."""
        if self._my_ids is None:
            pk_field = (
                self.queryset.model._meta.pk.attname  # pylint: disable=protected-access
            )
            self._my_ids = {o[pk_field] for o in data}

        return self._my_ids

    def merge_fk(self, data, field, prefetch):
        """Merge a foreign key."""
        # Strategy: pull out field_id values from each row, pass to
        #           prefetch queryset using `pk__in`.

        id_field = field.attname
        ids = {row[id_field] for row in data if id_field in row}
        prefetched_data = prefetch.query.get_ids(ids).execute()
        id_map = self._make_id_map(prefetched_data)

        for row in data:
            row[field.name] = id_map.get(row[id_field], None)

        return data

    def merge_o2o(self, data, field, prefetch):
        """Merge a one-to-one."""
        # Same as FK.
        return self.merge_fk(data, field, prefetch)

    def merge_o2or(self, data, field, prefetch, m2o_mode=False):
        """Merge a one-to-one remote."""
        # Strategy: get my IDs, filter remote model for rows pointing at
        #           my IDs.
        #           For m2o_mode, account for there many objects, while
        #           for o2or only support one reverse object.

        my_ids = self._get_my_ids(data)

        # If prefetching User.profile, construct filter like:
        #   Profile.objects.filter(user__in=<user_ids>)
        remote_field = field.remote_field.attname
        remote_filter_key = f"{remote_field}__in"
        filter_args = {remote_filter_key: my_ids}

        # Fetch remote objects
        remote_objects = prefetch.query.filter(**filter_args).execute()
        id_map = self._make_id_map(data, pk_field=self.pk_field)

        field_name = prefetch.field
        reverse_found = []  # IDs of local objects that were reversed
        for remote_obj in remote_objects:
            # Pull out ref on remote object pointing at us, and
            # get local object. There *should* always be a matching
            # local object because the remote objects were filtered
            # for those that referenced the local IDs.
            reverse_ref = remote_obj[remote_field]
            local_obj = id_map[reverse_ref]

            if m2o_mode:
                # in many-to-one mode, this is a list
                if field_name not in local_obj:
                    local_obj[field_name] = FastList([])
                local_obj[field_name].append(remote_obj)
            else:
                # in o2or mode, there can only be one
                local_obj[field_name] = remote_obj

            reverse_found.append(reverse_ref)

        # Set value to None for objects that didn't have a matching prefetch
        not_found = my_ids - set(reverse_found)
        for pk in not_found:
            id_map[pk][field_name] = FastList([]) if m2o_mode else None

        return data

    def merge_m2m(self, data, field, prefetch):
        """Merge a many-to-many."""
        # Strategy: pull out all my IDs, do a reverse filter on remote object.
        # e.g.: If prefetching User.groups, do
        #       Groups.filter(users__in=<user_ids>)

        my_ids = self._get_my_ids(data)

        base_qs = prefetch.query.queryset  # base queryset on remote model
        # get pk field name
        remote_pk_field = (
            base_qs.model._meta.pk.attname  # pylint: disable=protected-access
        )
        reverse_field = field.remote_field.name

        if reverse_field is None:
            # Note: We can't just reuse self.queryset here because it's
            #       been sliced already.
            filters = {f"{field.attname}__isnull": False}
            qs = self.queryset.model.objects.filter(pk__in=my_ids, **filters)
            joins = list(qs.values_list(field.attname, self.pk_field))
        else:
            # Get reverse mapping (for User.groups, get Group.users)
            # Note: `qs` already has base filter applied on remote model.
            filters = {f"{reverse_field}__in": my_ids}
            joins = list(
                base_qs.filter(**filters).values_list(remote_pk_field, reverse_field)
            )

        # Fetch remote objects, as values.
        remote_ids = {o[0] for o in joins}
        remote_objects = prefetch.query.get_ids(remote_ids).execute()
        id_map = self._make_id_map(remote_objects, pk_field=remote_pk_field)

        # Create mapping of local ID -> remote objects
        to_field = prefetch.field
        object_map = defaultdict(FastList)
        for remote_id, local_id in joins:
            if remote_id in id_map:
                object_map[local_id].append(id_map[remote_id])

        # Merge into working data set.
        for row in data:
            row[to_field] = object_map[row[self.pk_field]]

        return data

    def merge_m2o(self, data, field, prefetch):
        """Merge a many-to-one."""
        # Same as o2or but allow for many reverse objects.
        return self.merge_o2or(data, field, prefetch, m2o_mode=True)



"""Setup script for dynamic-rest-bse."""
from setuptools import find_packages, setup

NAME = "dynamic-rest-bse"
DESCRIPTION = "Dynamic API support to Django REST Framework. Forked..."
URL = "http://github.com/BillSchumacher/dynamic-rest"
VERSION = "2.4.3"
SCRIPTS = ["manage.py"]

with open("install_requires.txt", encoding="utf-8") as fp:
    INSTALL_REQUIRES = fp.readlines()

with open("README.rst", encoding="utf-8") as fp:
    LONG_DESCRIPTION = fp.read()

setup(
    description=DESCRIPTION,
    include_package_data=True,
    install_requires=INSTALL_REQUIRES,
    long_description=LONG_DESCRIPTION,
    long_description_content_type="text/x-rst",
    name=NAME,
    packages=find_packages(exclude=["benchmarks", "tests"]),
    scripts=SCRIPTS,
    url=URL,
    version=VERSION,
    classifiers=[
        "Framework :: Django",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: BSD License",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Topic :: Software Development :: Libraries :: Python Modules",
    ],
)



"""Paginator that supports dynamic page sizes and excludes count queries."""
# adapted from Django's django.core.paginator (3.2+ compatible)
# adds support for the "exclude_count" parameter

import inspect
from math import ceil

from django.core.paginator import EmptyPage, PageNotAnInteger, Paginator
from django.utils.functional import cached_property
from django.utils.inspect import method_has_no_args
from django.utils.translation import gettext_lazy as _


class DynamicPaginator(Paginator):
    """A subclass of Paginator that supports dynamic page sizes."""

    def __init__(self, *args, **kwargs):
        """Initialise the DynamicPaginator."""
        self.exclude_count = kwargs.pop("exclude_count", False)
        super().__init__(*args, **kwargs)

    def validate_number(self, number):
        """Validate the given 1-based page number."""
        try:
            number = int(number)
        except (TypeError, ValueError) as exc:
            raise PageNotAnInteger(_("That page number is not an integer")) from exc
        if number < 1:
            raise EmptyPage(_("That page number is less than 1"))
        if self.exclude_count:
            # skip validating against num_pages
            return number
        if number > self.num_pages:
            if number != 1 or not self.allow_empty_first_page:
                raise EmptyPage(_("That page contains no results"))
        return number

    def page(self, number):
        """Return a Page object for the given 1-based page number."""
        number = self.validate_number(number)
        per_page = self.per_page
        count = self.count
        bottom = (number - 1) * per_page
        top = bottom + per_page
        if self.exclude_count:
            # always fetch one extra item
            # to determine if more pages are available
            # and skip validation against count
            top += 1
        elif top + self.orphans >= count:
            top = count
        return self._get_page(self.object_list[bottom:top], number, self)

    @cached_property
    def count(self):
        """Return the total number of objects, across all pages."""
        if self.exclude_count:
            # always return 0, count should not be called
            return 0

        c = getattr(self.object_list, "count", None)
        if callable(c) and not inspect.isbuiltin(c) and method_has_no_args(c):
            return c()
        return len(self.object_list)

    @cached_property
    def num_pages(self):
        """Return the total number of pages."""
        if self.exclude_count:
            # always return 1, count should not be called
            return 1
        count = self.count
        if count == 0 and not self.allow_empty_first_page:
            return 0
        hits = max(1, count - self.orphans)
        return int(ceil(hits / float(self.per_page)))



"""This module contains custom renderer classes."""
from rest_framework.renderers import BrowsableAPIRenderer


class DynamicBrowsableAPIRenderer(BrowsableAPIRenderer):
    """Renderer class that adds directory support to the Browsable API."""

    template = "dynamic_rest/api.html"

    def get_context(self, data, accepted_media_type, renderer_context):
        """Return the context."""
        from dynamic_rest.routers import (  # pylint: disable=import-outside-toplevel
            get_directory,
        )

        context = super().get_context(data, accepted_media_type, renderer_context)
        request = context["request"]
        context["directory"] = get_directory(request)
        return context



"""This module contains custom router classes."""
import copy
import traceback
from collections import OrderedDict, defaultdict

import rest_framework
from django.urls import get_script_prefix
from rest_framework import views
from rest_framework.response import Response
from rest_framework.reverse import reverse
from rest_framework.routers import DefaultRouter, Route

from dynamic_rest.conf import settings
from dynamic_rest.meta import get_model_table


def replace_methodname(format_string, methodname):
    """Replace methodname in a format_string.

    Partially format a format_string, swapping out any
    '{methodname}' or '{methodnamehyphen}' components.
    """
    methodnamehyphen = methodname.replace("_", "-")
    ret = format_string
    ret = ret.replace("{methodname}", methodname)
    ret = ret.replace("{methodnamehyphen}", methodnamehyphen)
    return ret


directory = defaultdict(lambda: defaultdict(dict))
resource_map = {}
resource_name_map = {}
drf_version = tuple(int(part) for part in rest_framework.__version__.split("."))


def get_url(url, request):
    """Return a URL, or None if it is None."""
    return reverse(url, request=request) if url else url


def is_active_url(path, url):
    """Return True if the path starts with the URL."""
    return path.startswith(url) if url and path else False


def sort_key(r):
    """Return the first element of a tuple."""
    return r[0]


def get_directory(request):
    """Get API directory as a nested list of lists."""
    path = request.path
    directory_list = []

    # TODO(ant): support arbitrarily nested
    # structure, for now it is capped at a single level
    # for UX reasons
    for group_name, endpoints in sorted(directory.items(), key=sort_key):
        endpoints_list = []
        for endpoint_name, endpoint in sorted(endpoints.items(), key=sort_key):
            if endpoint_name[:1] == "_":
                continue
            endpoint_url = get_url(endpoint.get("_url", None), request)
            active = is_active_url(path, endpoint_url)
            endpoints_list.append((endpoint_name, endpoint_url, [], active))

        url = get_url(endpoints.get("_url", None), request)
        active = is_active_url(path, url)
        directory_list.append((group_name, url, endpoints_list, active))
    return directory_list


def modify_list_route(routes):
    """Modify the list route to allow bulk PATCH requests."""
    # Identify the list route, add PATCH, so we can
    # have bulk PATCH requests
    if not settings.ENABLE_BULK_UPDATE:
        return
    list_route = next(i for i in routes if i.name == "{basename}-list")
    list_route.mapping["patch"] = "partial_update"
    list_route.mapping["delete"] = "destroy"


class DynamicRouter(DefaultRouter):
    """A router that generates a global API directory."""

    routes = copy.deepcopy(DefaultRouter.routes)
    modify_list_route(routes)

    def __init__(self, *args, **kwargs):
        """Initialise the DynamicRouter."""
        optional_trailing_slash = kwargs.pop("optional_trailing_slash", True)
        super().__init__(*args, **kwargs)
        if optional_trailing_slash:
            self.trailing_slash = "/?"

    def get_api_root_view(self, **__):
        """Return API root view, using the global directory."""

        class API(views.APIView):
            """API root view, using the global directory."""

            _ignore_model_permissions = True

            def get(self, request, *_, **__):
                """Return the API directory."""
                directory_list = get_directory(request)
                result = OrderedDict()
                for group_name, url, endpoints, _ in directory_list:
                    if url:
                        result[group_name] = url
                    else:
                        group = OrderedDict()
                        for endpoint_name, url, _, _ in endpoints:
                            group[endpoint_name] = url
                        result[group_name] = group
                return Response(result)

        return API.as_view()

    def register(self, prefix, viewset, basename=None):
        """Add any registered route into a global API directory.

        If the prefix includes a path separator,
        store the URL in the directory under the first path segment.
        Otherwise, store it as-is.

        For example, if there are two registered prefixes,
        'v1/users' and 'groups', `directory` will look liks:

        {
            'v1': {
                'users': {
                    '_url': 'users-list'
                    '_viewset': <class 'UserViewSet'>
                },
            }
            'groups': {
               '_url': 'groups-list'
               '_viewset': <class 'GroupViewSet'>
            }
        }
        """
        if basename is None:
            basename = prefix

        super().register(prefix, viewset, basename)

        prefix_parts = prefix.split("/")
        if len(prefix_parts) > 1:
            prefix = prefix_parts[0]
            endpoint = "/".join(prefix_parts[1:])
        else:
            endpoint = prefix
            prefix = None

        current = directory[prefix] if prefix else directory.get(prefix, directory)
        list_name = self.routes[0].name
        url_name = list_name.format(basename=basename)
        current[endpoint]["_url"] = url_name
        current[endpoint]["_viewset"] = viewset

    def register_resource(self, viewset, namespace=None):
        """Register a resource.

        Register a viewset that should be considered the canonical
        endpoint for a particular resource. In addition to generating
        and registering the route, it adds the route in a reverse map
        to allow DREST to build the canonical URL for a given resource.

        Arguments:
            viewset: A viewset class.
            namespace: (Optional) A namespace to prepend to the URL.
        """
        # Try to extract resource name from viewset.
        try:
            serializer = viewset.serializer_class()
            resource_key = serializer.get_resource_key()
            resource_name = serializer.get_name()
            path_name = serializer.get_plural_name()
        except BaseException as exc:
            traceback.print_exc()
            raise RuntimeError(
                f"Failed to extract resource name from viewset: '{viewset}'."
                " It, or its serializer, may not be DREST-compatible."
            ) from exc

        # Construct canonical path and register it.
        if namespace:
            namespace = namespace.rstrip("/") + "/"
        base_path = namespace or ""
        base_path = f"{base_path}{path_name}"
        self.register(base_path, viewset)

        # Make sure resource isn't already registered.
        if resource_key in resource_map:
            raise RuntimeError(
                f"The resource '{resource_key}' has already been"
                f" mapped to '{resource_map[resource_key]['path']}'."
                " Each resource can only be mapped to one canonical"
                " path. "
            )

        # Register resource in reverse map.
        resource_map[resource_key] = {"path": base_path, "viewset": viewset}

        # Make sure the resource name isn't registered, either
        # TODO: Think of a better way to clean this up, there's a lot of
        # duplicated effort here, between `resource_name` and `resource_key`
        # This resource name -> key mapping is currently only used by
        # the DynamicGenericRelationField
        if resource_name in resource_name_map:
            resource_key = resource_name_map[resource_name]
            raise RuntimeError(
                f"The resource name '{resource_name}' has already been mapped"
                f" to '{resource_map[resource_key]['path']}'."
                " A resource name can only be used once."
            )

        # map the resource name to the resource key for easier lookup
        resource_name_map[resource_name] = resource_key

    @staticmethod
    def get_canonical_path(resource_key, pk=None):
        """Return canonical resource path.

        Arguments:
            resource_key: Canonical resource key.
                i.e. Serializer.get_resource_key().
            pk: (Optional) Object's primary key for a single-resource URL.
        Returns: Absolute URL as string.
        """
        if resource_key not in resource_map:
            # Note: Maybe raise?
            return None

        base_path = get_script_prefix() + resource_map[resource_key]["path"]
        return f"{base_path}/{pk}/" if pk else base_path

    @staticmethod
    def get_canonical_serializer(
        resource_key, model=None, instance=None, resource_name=None
    ):
        """Return canonical serializer for a given resource name.

        Arguments:
            resource_key: Resource key, usually DB table for model-based
                resources, otherwise the plural name.
            model: (Optional) Model class to look up by.
            instance: (Optional) Model object instance.
            resource_name: The resource name.
        Returns: serializer class
        """
        if model:
            resource_key = get_model_table(model)
        elif instance:
            resource_key = instance._meta.db_table  # pylint: disable=protected-access
        elif resource_name:
            resource_key = resource_name_map[resource_name]

        if resource_key not in resource_map:
            return None

        return resource_map[resource_key]["viewset"].serializer_class

    def get_routes(self, viewset):
        """Get routes for a viewset.

        DREST routes injection, overrides DRF's get_routes() method, which
        gets called for each registered viewset.
        """
        routes = super().get_routes(viewset)
        routes += self.get_relation_routes(viewset)
        return routes

    @staticmethod
    def get_relation_routes(viewset):
        """Get routes for a viewset's relational fields.

        Generate routes to serve relational objects. This method will add
        a sub-URL for each relational field.

        e.g.
        A viewset for the following serializer:

          class UserSerializer(..):
              events = DynamicRelationField(EventSerializer, many=True)
              groups = DynamicRelationField(GroupSerializer, many=True)
              location = DynamicRelationField(LocationSerializer)

        will have the following URLs added:

          /users/<pk>/events/
          /users/<pk>/groups/
          /users/<pk>/location/
        """
        routes = []

        if not hasattr(viewset, "serializer_class"):
            return routes
        if not hasattr(viewset, "list_related"):
            return routes

        serializer = viewset.serializer_class()
        # TODO: Is this really a list? Seems like it should be dict.
        fields = getattr(serializer, "get_link_fields", lambda: [])()

        route_name = "{basename}-{methodnamehyphen}"
        route_compat_kwargs = {"detail": False} if drf_version >= (3, 8, 0) else {}
        methodname = "list_related"
        for field_name, _ in fields.items():
            url = (
                r"^{prefix}/{lookup}/(?P<field_name>%s)"
                "{trailing_slash}$" % field_name
            )
            routes.append(
                Route(
                    url=url,
                    mapping={"get": methodname},
                    name=replace_methodname(route_name, field_name),
                    initkwargs={},
                    **route_compat_kwargs,
                )
            )
        return routes



"""Tests for the dynamic_rest.routers module."""
import os

from django.urls import clear_script_prefix, set_script_prefix
from rest_framework.routers import DefaultRouter

from dynamic_rest.meta import get_model_table
from dynamic_rest.routers import DynamicRouter, Route
from tests.models import Dog
from tests.serializers import CatSerializer, DogSerializer
from tests.urls import urlpatterns  # noqa pylint: disable=unused-import

if os.getenv("DATABASE_URL"):
    from tests.test_cases import ResetAPITestCase as TestCase
else:
    from tests.test_cases import APITestCase as TestCase


class TestDynamicRouter(TestCase):
    """Test case for dynamic_rest.routers."""

    def test_get_canonical_path(self):
        """Test get canonical path."""
        rsrc_key = DogSerializer().get_resource_key()
        self.assertEqual("/dogs", DynamicRouter.get_canonical_path(rsrc_key))

    def test_get_canonical_path_with_prefix(self):
        """Test get canonical path with prefix."""
        set_script_prefix("/v2/")
        rsrc_key = DogSerializer().get_resource_key()
        self.assertEqual("/v2/dogs", DynamicRouter.get_canonical_path(rsrc_key))
        clear_script_prefix()

    def test_get_canonical_path_with_pk(self):
        """Test get canonical path with pk."""
        rsrc_key = DogSerializer().get_resource_key()
        self.assertEqual("/dogs/1/", DynamicRouter.get_canonical_path(rsrc_key, pk="1"))

    def test_get_canonical_path_with_keyspace(self):
        """Test get canonical path with keyspace."""
        rsrc_key = CatSerializer().get_resource_key()
        self.assertEqual("/v2/cats", DynamicRouter.get_canonical_path(rsrc_key))

    def test_get_canonical_serializer(self):
        """Test get canonical serializer."""
        rsrc_key = get_model_table(Dog)
        self.assertEqual(
            DogSerializer, DynamicRouter.get_canonical_serializer(rsrc_key)
        )

    def test_get_canonical_serializer_by_model(self):
        """Test get canonical serializer by model."""
        self.assertEqual(
            DogSerializer, DynamicRouter.get_canonical_serializer(None, model=Dog)
        )

    def test_get_canonical_serializer_by_instance(self):
        """Test get canonical serializer by instance."""
        dog = Dog.objects.create(name="Snoopy", fur_color="black and white", origin="")
        self.assertEqual(
            DogSerializer, DynamicRouter.get_canonical_serializer(None, instance=dog)
        )

    def test_rest_framework_router_unmodified(self):
        """Test rest framework router unmodified."""
        if hasattr(self, "assertCountEqual"):
            method = self.assertCountEqual
        else:
            method = self.assertItemsEqual

        method(
            [
                {"post": "create", "get": "list"},
                {
                    "put": "update",
                    "patch": "partial_update",
                    "delete": "destroy",
                    "get": "retrieve",
                },
            ],
            [
                route.mapping
                for route in DefaultRouter.routes
                if isinstance(route, Route)
            ],
        )



"""Benchmark setup for DREST."""
from benchmarks_app.models import Group, Permission, User
from dynamic_rest import fields, routers, serializers, viewsets


class UserSerializer(serializers.DynamicModelSerializer):
    """User serializer."""

    class Meta:
        """Meta class."""

        model = User
        name = "user"
        fields = ("id", "name", "groups")

    groups = fields.DynamicRelationField(
        "GroupSerializer", embed=True, many=True, deferred=True
    )


class GroupSerializer(serializers.DynamicModelSerializer):
    """Group serializer."""

    class Meta:
        """Meta class."""

        model = Group
        name = "group"
        fields = ("id", "name", "permissions")

    permissions = fields.DynamicRelationField(
        "PermissionSerializer", embed=True, many=True, deferred=True
    )


class PermissionSerializer(serializers.DynamicModelSerializer):
    """Permission serializer."""

    class Meta:
        """Meta class."""

        model = Permission
        name = "permission"
        fields = ("id", "name")


# DREST views


class UserViewSet(viewsets.DynamicModelViewSet):
    """User viewset."""

    queryset = User.objects.all()
    serializer_class = UserSerializer


# DREST router

router = routers.DynamicRouter()
router.register(r"drest/users", UserViewSet)
